"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class Code(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _Type:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _TypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[Code._Type.ValueType], builtins.type):  # noqa: F821
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        NOT_SET: Code._Type.ValueType  # 0
        NBOX: Code._Type.ValueType  # 1
        ZIP: Code._Type.ValueType  # 2
        BINARY: Code._Type.ValueType  # 3

    class Type(_Type, metaclass=_TypeEnumTypeWrapper): ...
    NOT_SET: Code.Type.ValueType  # 0
    NBOX: Code.Type.ValueType  # 1
    ZIP: Code.Type.ValueType  # 2
    BINARY: Code.Type.ValueType  # 3

    @typing_extensions.final
    class S3MetaEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.str
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    S3_URL_FIELD_NUMBER: builtins.int
    SIZE_FIELD_NUMBER: builtins.int
    TYPE_FIELD_NUMBER: builtins.int
    S3_META_FIELD_NUMBER: builtins.int
    s3_url: builtins.str
    size: builtins.int
    type: global___Code.Type.ValueType
    @property
    def s3_meta(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]: ...
    def __init__(
        self,
        *,
        s3_url: builtins.str = ...,
        size: builtins.int = ...,
        type: global___Code.Type.ValueType = ...,
        s3_meta: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["s3_meta", b"s3_meta", "s3_url", b"s3_url", "size", b"size", "type", b"type"]) -> None: ...

global___Code = Code

@typing_extensions.final
class NBXAuthInfo(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    USERNAME_FIELD_NUMBER: builtins.int
    WORKSPACE_ID_FIELD_NUMBER: builtins.int
    ACCESS_TOKEN_FIELD_NUMBER: builtins.int
    REFRESH_TOKEN_FIELD_NUMBER: builtins.int
    ACCESS_KEY_FIELD_NUMBER: builtins.int
    username: builtins.str
    workspace_id: builtins.str
    """Workspace ID"""
    access_token: builtins.str
    refresh_token: builtins.str
    access_key: builtins.str
    """NBX Access Key for use with Jobs"""
    def __init__(
        self,
        *,
        username: builtins.str = ...,
        workspace_id: builtins.str = ...,
        access_token: builtins.str = ...,
        refresh_token: builtins.str = ...,
        access_key: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["access_key", b"access_key", "access_token", b"access_token", "refresh_token", b"refresh_token", "username", b"username", "workspace_id", b"workspace_id"]) -> None: ...

global___NBXAuthInfo = NBXAuthInfo

@typing_extensions.final
class Resource(google.protobuf.message.Message):
    """Resources required by Job"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CPU_FIELD_NUMBER: builtins.int
    MEMORY_FIELD_NUMBER: builtins.int
    GPU_FIELD_NUMBER: builtins.int
    GPU_COUNT_FIELD_NUMBER: builtins.int
    DISK_SIZE_FIELD_NUMBER: builtins.int
    TIMEOUT_FIELD_NUMBER: builtins.int
    MAX_RETRIES_FIELD_NUMBER: builtins.int
    cpu: builtins.str
    """CPU in milli-core"""
    memory: builtins.str
    """Memory in MiB"""
    gpu: builtins.str
    """GPU Type"""
    gpu_count: builtins.str
    """Number of GPUs as a fraction"""
    disk_size: builtins.str
    """in GiBs"""
    timeout: builtins.int
    """in seconds. Note this is inclusive of jobs and init containers, enforced by k8s, regardless of `max_retries` set"""
    max_retries: builtins.int
    """max number of retries for a job"""
    def __init__(
        self,
        *,
        cpu: builtins.str = ...,
        memory: builtins.str = ...,
        gpu: builtins.str = ...,
        gpu_count: builtins.str = ...,
        disk_size: builtins.str = ...,
        timeout: builtins.int = ...,
        max_retries: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["cpu", b"cpu", "disk_size", b"disk_size", "gpu", b"gpu", "gpu_count", b"gpu_count", "max_retries", b"max_retries", "memory", b"memory", "timeout", b"timeout"]) -> None: ...

global___Resource = Resource



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.message
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class Acknowledge(google.protobuf.message.Message):
    """generic response to any message"""
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    SUCCESS_FIELD_NUMBER: builtins.int
    MESSAGE_FIELD_NUMBER: builtins.int
    success: builtins.bool
    message: typing.Text
    """message can contain message or traceback if not success"""

    def __init__(self,
        *,
        success: builtins.bool = ...,
        message: typing.Text = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["message",b"message","success",b"success"]) -> None: ...
global___Acknowledge = Acknowledge



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: serve.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from google.protobuf import field_mask_pb2 as google_dot_protobuf_dot_field__mask__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from . import common_pb2 as common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0bserve.proto\x12\x06\x64\x65ploy\x1a\x1bgoogle/protobuf/empty.proto\x1a google/protobuf/field_mask.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x0c\x63ommon.proto\"\x96\x03\n\x07Serving\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x1d\n\x06models\x18\x03 \x03(\x0b\x32\r.deploy.Model\x12-\n\x13\x64\x65ployment_endpoint\x18\x04 \x01(\x0b\x32\x10.deploy.Endpoint\x12)\n\x0fmodel_endpoints\x18\x05 \x03(\x0b\x32\x10.deploy.Endpoint\x12\"\n\x08resource\x18\x06 \x01(\x0b\x32\x10.common.Resource\x12-\n\x10\x63pu_scale_config\x18\x07 \x01(\x0b\x32\x13.deploy.ScaleConfig\x12\x13\n\x0b\x64\x65scription\x18\x08 \x01(\t\x12/\n\x08metadata\x18\t \x03(\x0b\x32\x1d.deploy.Serving.MetadataEntry\x12.\n\ncreated_at\x18\n \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x1a/\n\rMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"7\n\x08\x45ndpoint\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0b\n\x03url\x18\x02 \x01(\t\x12\x12\n\naccess_key\x18\x03 \x01(\t\"\xd6\x01\n\rServingStatus\x12\x33\n\rupload_status\x18\x01 \x01(\x0e\x32\x1c.deploy.ServingStatus.Status\x12\x33\n\rdeploy_status\x18\x02 \x01(\x0e\x32\x1c.deploy.ServingStatus.Status\"[\n\x06Status\x12\x16\n\x12STATUS_UNSPECIFIED\x10\x00\x12\x12\n\x0eSTATUS_PENDING\x10\x01\x12\x12\n\x0eSTATUS_SUCCESS\x10\x02\x12\x11\n\rSTATUS_FAILED\x10\x03\"L\n\x0bScaleConfig\x12\x14\n\x0cmin_replicas\x18\x01 \x01(\x05\x12\x14\n\x0cmax_replicas\x18\x02 \x01(\x05\x12\x11\n\tthreshold\x18\x03 \x01(\x02\"Z\n\x0eServingRequest\x12&\n\tauth_info\x18\x01 \x01(\x0b\x32\x13.common.NBXAuthInfo\x12 \n\x07serving\x18\x02 \x01(\x0b\x32\x0f.deploy.Serving\"\x91\x01\n\x14UpdateServingRequest\x12&\n\tauth_info\x18\x01 \x01(\x0b\x32\x13.common.NBXAuthInfo\x12 \n\x07serving\x18\x02 \x01(\x0b\x32\x0f.deploy.Serving\x12/\n\x0bupdate_mask\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.FieldMask\"[\n\x12ServingListRequest\x12&\n\tauth_info\x18\x01 \x01(\x0b\x32\x13.common.NBXAuthInfo\x12\r\n\x05limit\x18\x02 \x01(\x05\x12\x0e\n\x06offset\x18\x03 \x01(\x05\"K\n\x13ServingListResponse\x12!\n\x08servings\x18\x01 \x03(\x0b\x32\x0f.deploy.Serving\x12\x11\n\tnext_page\x18\x02 \x01(\x05\"\xcc\x04\n\x05Model\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x18\n\x10serving_group_id\x18\x03 \x01(\t\x12.\n\ncreated_at\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\'\n\x04type\x18\x05 \x01(\x0e\x32\x19.deploy.Model.ServingType\x12\"\n\x08resource\x18\x06 \x01(\x0b\x32\x10.common.Resource\x12\x1a\n\x04\x63ode\x18\x07 \x01(\x0b\x32\x0c.common.Code\x12\"\n\x08\x65ndpoint\x18\x08 \x01(\x0b\x32\x10.deploy.Endpoint\x12+\n\npin_status\x18\t \x01(\x0e\x32\x17.deploy.Model.PinStatus\x12-\n\x08metadata\x18\n \x03(\x0b\x32\x1b.deploy.Model.MetadataEntry\x12%\n\x06status\x18\x0e \x01(\x0b\x32\x15.deploy.ServingStatus\x1a/\n\rMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"E\n\x0bServingType\x12\x1c\n\x18SERVING_TYPE_UNSPECIFIED\x10\x00\x12\x18\n\x14SERVING_TYPE_NBOX_OP\x10\x01\"W\n\tPinStatus\x12\x1a\n\x16PIN_STATUS_UNSPECIFIED\x10\x00\x12\x15\n\x11PIN_STATUS_PINNED\x10\x01\x12\x17\n\x13PIN_STATUS_UNPINNED\x10\x02\"T\n\x0cModelRequest\x12&\n\tauth_info\x18\x01 \x01(\x0b\x32\x13.common.NBXAuthInfo\x12\x1c\n\x05model\x18\x02 \x01(\x0b\x32\r.deploy.Model\"m\n\x10ModelListRequest\x12&\n\tauth_info\x18\x01 \x01(\x0b\x32\x13.common.NBXAuthInfo\x12\x12\n\nserving_id\x18\x02 \x01(\t\x12\r\n\x05limit\x18\x03 \x01(\x05\x12\x0e\n\x06offset\x18\x04 \x01(\x05\"E\n\x11ModelListResponse\x12\x1d\n\x06models\x18\x01 \x03(\x0b\x32\r.deploy.Model\x12\x11\n\tnext_page\x18\x02 \x01(\x05\"$\n\x08ModelLog\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0c\n\x04logs\x18\x02 \x03(\t2\xc4\x02\n\x0eServingService\x12\x31\n\x06\x43reate\x12\x16.deploy.ServingRequest\x1a\x0f.deploy.Serving\x12\x35\n\nGetServing\x12\x16.deploy.ServingRequest\x1a\x0f.deploy.Serving\x12G\n\x0cListServings\x12\x1a.deploy.ServingListRequest\x1a\x1b.deploy.ServingListResponse\x12?\n\rDeleteServing\x12\x16.deploy.ServingRequest\x1a\x16.google.protobuf.Empty\x12>\n\rUpdateServing\x12\x1c.deploy.UpdateServingRequest\x1a\x0f.deploy.Serving2\x9d\x03\n\x0cModelService\x12\x36\n\x05Watch\x12\x14.deploy.ModelRequest\x1a\x15.deploy.ServingStatus0\x01\x12\x32\n\x0bUploadModel\x12\x14.deploy.ModelRequest\x1a\r.deploy.Model\x12-\n\x06\x44\x65ploy\x12\x14.deploy.ModelRequest\x1a\r.deploy.Model\x12\x35\n\tModelLogs\x12\x14.deploy.ModelRequest\x1a\x10.deploy.ModelLog0\x01\x12\x41\n\nListModels\x12\x18.deploy.ModelListRequest\x1a\x19.deploy.ModelListResponse\x12;\n\x0b\x44\x65leteModel\x12\x14.deploy.ModelRequest\x1a\x16.google.protobuf.Empty\x12;\n\x0bSetModelPin\x12\x14.deploy.ModelRequest\x1a\x16.google.protobuf.EmptyB4Z2github.com/NimbleBoxAI/jobs-architecture/deploy_pbb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'serve_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z2github.com/NimbleBoxAI/jobs-architecture/deploy_pb'
  _SERVING_METADATAENTRY._options = None
  _SERVING_METADATAENTRY._serialized_options = b'8\001'
  _MODEL_METADATAENTRY._options = None
  _MODEL_METADATAENTRY._serialized_options = b'8\001'
  _SERVING._serialized_start=134
  _SERVING._serialized_end=540
  _SERVING_METADATAENTRY._serialized_start=493
  _SERVING_METADATAENTRY._serialized_end=540
  _ENDPOINT._serialized_start=542
  _ENDPOINT._serialized_end=597
  _SERVINGSTATUS._serialized_start=600
  _SERVINGSTATUS._serialized_end=814
  _SERVINGSTATUS_STATUS._serialized_start=723
  _SERVINGSTATUS_STATUS._serialized_end=814
  _SCALECONFIG._serialized_start=816
  _SCALECONFIG._serialized_end=892
  _SERVINGREQUEST._serialized_start=894
  _SERVINGREQUEST._serialized_end=984
  _UPDATESERVINGREQUEST._serialized_start=987
  _UPDATESERVINGREQUEST._serialized_end=1132
  _SERVINGLISTREQUEST._serialized_start=1134
  _SERVINGLISTREQUEST._serialized_end=1225
  _SERVINGLISTRESPONSE._serialized_start=1227
  _SERVINGLISTRESPONSE._serialized_end=1302
  _MODEL._serialized_start=1305
  _MODEL._serialized_end=1893
  _MODEL_METADATAENTRY._serialized_start=493
  _MODEL_METADATAENTRY._serialized_end=540
  _MODEL_SERVINGTYPE._serialized_start=1735
  _MODEL_SERVINGTYPE._serialized_end=1804
  _MODEL_PINSTATUS._serialized_start=1806
  _MODEL_PINSTATUS._serialized_end=1893
  _MODELREQUEST._serialized_start=1895
  _MODELREQUEST._serialized_end=1979
  _MODELLISTREQUEST._serialized_start=1981
  _MODELLISTREQUEST._serialized_end=2090
  _MODELLISTRESPONSE._serialized_start=2092
  _MODELLISTRESPONSE._serialized_end=2161
  _MODELLOG._serialized_start=2163
  _MODELLOG._serialized_end=2199
  _SERVINGSERVICE._serialized_start=2202
  _SERVINGSERVICE._serialized_end=2526
  _MODELSERVICE._serialized_start=2529
  _MODELSERVICE._serialized_end=2942
# @@protoc_insertion_point(module_scope)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class Relic(google.protobuf.message.Message):
    """Some terms and glossary for sanity:
    Relic: Relic is a folder on the cloud like a router which links to different files.
        A relic is the tree-group of individual files.
    RelicFile: RelicFile is a file in a relic which points to the correct location of the
        file on the cloud. Since this is not an object store in itself and only a pointer,
        different clients are responsible for figuring out how to download/uplaod the
        files.

    the main outer thing responsible for grouping files together
    """
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    ID_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    CREATED_ON_FIELD_NUMBER: builtins.int
    LAST_MODIFIED_FIELD_NUMBER: builtins.int
    STARRED_FIELD_NUMBER: builtins.int
    TAGS_FIELD_NUMBER: builtins.int
    WORKSPACE_ID_FIELD_NUMBER: builtins.int
    BUCKET_META_FIELD_NUMBER: builtins.int
    id: typing.Text
    """these are the primary fields"""

    name: typing.Text
    created_on: builtins.int
    last_modified: builtins.int
    starred: builtins.bool
    """this is the starred functionality in NBX-Relics"""

    @property
    def tags(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text]:
        """tags are like labels in NBX-Relics"""
        pass
    workspace_id: typing.Text
    """this is the workspace id"""

    @property
    def bucket_meta(self) -> global___BucketMetadata:
        """there are other things that may not be needed by the FE, think of this as the 
        NOTE: these are high index values so they can be expanded in the 1xx range
        a world with 100 clouds supported, would love to see that day.
        this can potentially become the metadata of the relic
        """
        pass
    def __init__(self,
        *,
        id: typing.Text = ...,
        name: typing.Text = ...,
        created_on: builtins.int = ...,
        last_modified: builtins.int = ...,
        starred: builtins.bool = ...,
        tags: typing.Optional[typing.Iterable[typing.Text]] = ...,
        workspace_id: typing.Text = ...,
        bucket_meta: typing.Optional[global___BucketMetadata] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["bucket_meta",b"bucket_meta"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["bucket_meta",b"bucket_meta","created_on",b"created_on","id",b"id","last_modified",b"last_modified","name",b"name","starred",b"starred","tags",b"tags","workspace_id",b"workspace_id"]) -> None: ...
global___Relic = Relic

class BucketMetadata(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    class _Backend:
        ValueType = typing.NewType('ValueType', builtins.int)
        V: typing_extensions.TypeAlias = ValueType
    class _BackendEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[BucketMetadata._Backend.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        UNSET: BucketMetadata._Backend.ValueType  # 0
        """leave it to NBX to figure out the backend"""

        AWS_S3: BucketMetadata._Backend.ValueType  # 1
        GCP_GCS: BucketMetadata._Backend.ValueType  # 2
        AZURE_BLOB: BucketMetadata._Backend.ValueType  # 3
        OCI_OB: BucketMetadata._Backend.ValueType  # 4
        DO_SPACES: BucketMetadata._Backend.ValueType  # 5
    class Backend(_Backend, metaclass=_BackendEnumTypeWrapper):
        pass

    UNSET: BucketMetadata.Backend.ValueType  # 0
    """leave it to NBX to figure out the backend"""

    AWS_S3: BucketMetadata.Backend.ValueType  # 1
    GCP_GCS: BucketMetadata.Backend.ValueType  # 2
    AZURE_BLOB: BucketMetadata.Backend.ValueType  # 3
    OCI_OB: BucketMetadata.Backend.ValueType  # 4
    DO_SPACES: BucketMetadata.Backend.ValueType  # 5

    class BucketTagsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text
        value: typing.Text
        def __init__(self,
            *,
            key: typing.Text = ...,
            value: typing.Text = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    BUCKET_NAME_FIELD_NUMBER: builtins.int
    REGION_FIELD_NUMBER: builtins.int
    BACKEND_FIELD_NUMBER: builtins.int
    BUCKET_TAGS_FIELD_NUMBER: builtins.int
    bucket_name: typing.Text
    region: typing.Text
    backend: global___BucketMetadata.Backend.ValueType
    @property
    def bucket_tags(self) -> google.protobuf.internal.containers.ScalarMap[typing.Text, typing.Text]: ...
    def __init__(self,
        *,
        bucket_name: typing.Text = ...,
        region: typing.Text = ...,
        backend: global___BucketMetadata.Backend.ValueType = ...,
        bucket_tags: typing.Optional[typing.Mapping[typing.Text, typing.Text]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["backend",b"backend","bucket_name",b"bucket_name","bucket_tags",b"bucket_tags","region",b"region"]) -> None: ...
global___BucketMetadata = BucketMetadata

class RelicFile(google.protobuf.message.Message):
    """RelicFile is like the individual object on the object store, at the end of
    the day an object store is nothing but a key value pair, where the "filepath"
    is the key and the "file" is the value. This basically means that "/" is
    ignored.
    Most of the fields are inspired from the macOS Finder.
    """
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    class _RelicType:
        ValueType = typing.NewType('ValueType', builtins.int)
        V: typing_extensions.TypeAlias = ValueType
    class _RelicTypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[RelicFile._RelicType.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        FILE: RelicFile._RelicType.ValueType  # 0
        """file icon"""

        FOLDER: RelicFile._RelicType.ValueType  # 1
        """folder icon"""

        RELIC: RelicFile._RelicType.ValueType  # 2
        """potentially in the future we can symlink things"""

    class RelicType(_RelicType, metaclass=_RelicTypeEnumTypeWrapper):
        pass

    FILE: RelicFile.RelicType.ValueType  # 0
    """file icon"""

    FOLDER: RelicFile.RelicType.ValueType  # 1
    """folder icon"""

    RELIC: RelicFile.RelicType.ValueType  # 2
    """potentially in the future we can symlink things"""


    class HeadersEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text
        value: typing.Text
        def __init__(self,
            *,
            key: typing.Text = ...,
            value: typing.Text = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    class BodyEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text
        value: typing.Text
        def __init__(self,
            *,
            key: typing.Text = ...,
            value: typing.Text = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    NAME_FIELD_NUMBER: builtins.int
    CREATED_ON_FIELD_NUMBER: builtins.int
    LAST_MODIFIED_FIELD_NUMBER: builtins.int
    STARRED_FIELD_NUMBER: builtins.int
    TAGS_FIELD_NUMBER: builtins.int
    SIZE_FIELD_NUMBER: builtins.int
    COMMENT_FIELD_NUMBER: builtins.int
    USERNAME_FIELD_NUMBER: builtins.int
    TYPE_FIELD_NUMBER: builtins.int
    WORKSPACE_ID_FIELD_NUMBER: builtins.int
    RELIC_NAME_FIELD_NUMBER: builtins.int
    RELIC_ID_FIELD_NUMBER: builtins.int
    URL_FIELD_NUMBER: builtins.int
    HEADERS_FIELD_NUMBER: builtins.int
    BODY_FIELD_NUMBER: builtins.int
    name: typing.Text
    created_on: builtins.int
    last_modified: builtins.int
    starred: builtins.bool
    """this is the starred functionality in NBX-Relics"""

    @property
    def tags(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text]:
        """tags are like labels in NBX-Relics"""
        pass
    size: builtins.int
    """the size of the file"""

    comment: typing.Text
    """There are other "Human" aspects like comments, etc. This is all part of collaboration"""

    username: typing.Text
    """the creator of the file, this is a little bit ambivalent. Is this file created by the
    user or by the job. For now we are saying that this is the user who created the file.
    """

    type: global___RelicFile.RelicType.ValueType
    workspace_id: typing.Text
    """this is the workspace id"""

    relic_name: typing.Text
    """the name of the parent relic"""

    relic_id: typing.Text
    """the name of the parent relic"""

    url: typing.Text
    """NOTE: these are high index values so they can be expanded in the 1xx range
    this is all for the different clouds
    the url of the file
    """

    @property
    def headers(self) -> google.protobuf.internal.containers.ScalarMap[typing.Text, typing.Text]:
        """the headers to be used by the client"""
        pass
    @property
    def body(self) -> google.protobuf.internal.containers.ScalarMap[typing.Text, typing.Text]:
        """the body to be used by the client"""
        pass
    def __init__(self,
        *,
        name: typing.Text = ...,
        created_on: builtins.int = ...,
        last_modified: builtins.int = ...,
        starred: builtins.bool = ...,
        tags: typing.Optional[typing.Iterable[typing.Text]] = ...,
        size: builtins.int = ...,
        comment: typing.Text = ...,
        username: typing.Text = ...,
        type: global___RelicFile.RelicType.ValueType = ...,
        workspace_id: typing.Text = ...,
        relic_name: typing.Text = ...,
        relic_id: typing.Text = ...,
        url: typing.Text = ...,
        headers: typing.Optional[typing.Mapping[typing.Text, typing.Text]] = ...,
        body: typing.Optional[typing.Mapping[typing.Text, typing.Text]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["body",b"body","comment",b"comment","created_on",b"created_on","headers",b"headers","last_modified",b"last_modified","name",b"name","relic_id",b"relic_id","relic_name",b"relic_name","size",b"size","starred",b"starred","tags",b"tags","type",b"type","url",b"url","username",b"username","workspace_id",b"workspace_id"]) -> None: ...
global___RelicFile = RelicFile



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import collections.abc
import common_pb2
import google.protobuf.descriptor
import google.protobuf.field_mask_pb2
import google.protobuf.internal.containers
import google.protobuf.message
import google.protobuf.timestamp_pb2
import job_pb2
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class JobRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    AUTH_INFO_FIELD_NUMBER: builtins.int
    JOB_FIELD_NUMBER: builtins.int
    @property
    def auth_info(self) -> common_pb2.NBXAuthInfo: ...
    @property
    def job(self) -> job_pb2.Job: ...
    def __init__(
        self,
        *,
        auth_info: common_pb2.NBXAuthInfo | None = ...,
        job: job_pb2.Job | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["auth_info", b"auth_info", "job", b"job"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["auth_info", b"auth_info", "job", b"job"]) -> None: ...

global___JobRequest = JobRequest

@typing_extensions.final
class UpdateJobRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    AUTH_INFO_FIELD_NUMBER: builtins.int
    JOB_FIELD_NUMBER: builtins.int
    UPDATE_MASK_FIELD_NUMBER: builtins.int
    @property
    def auth_info(self) -> common_pb2.NBXAuthInfo: ...
    @property
    def job(self) -> job_pb2.Job: ...
    @property
    def update_mask(self) -> google.protobuf.field_mask_pb2.FieldMask: ...
    def __init__(
        self,
        *,
        auth_info: common_pb2.NBXAuthInfo | None = ...,
        job: job_pb2.Job | None = ...,
        update_mask: google.protobuf.field_mask_pb2.FieldMask | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["auth_info", b"auth_info", "job", b"job", "update_mask", b"update_mask"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["auth_info", b"auth_info", "job", b"job", "update_mask", b"update_mask"]) -> None: ...

global___UpdateJobRequest = UpdateJobRequest

@typing_extensions.final
class ListJobsRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    AUTH_INFO_FIELD_NUMBER: builtins.int
    LIMIT_FIELD_NUMBER: builtins.int
    OFFSET_FIELD_NUMBER: builtins.int
    @property
    def auth_info(self) -> common_pb2.NBXAuthInfo: ...
    limit: builtins.int
    offset: builtins.int
    def __init__(
        self,
        *,
        auth_info: common_pb2.NBXAuthInfo | None = ...,
        limit: builtins.int = ...,
        offset: builtins.int = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["auth_info", b"auth_info"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["auth_info", b"auth_info", "limit", b"limit", "offset", b"offset"]) -> None: ...

global___ListJobsRequest = ListJobsRequest

@typing_extensions.final
class ListJobsResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    JOBS_FIELD_NUMBER: builtins.int
    NEXT_PAGE_FIELD_NUMBER: builtins.int
    @property
    def jobs(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[job_pb2.Job]: ...
    next_page: builtins.int
    def __init__(
        self,
        *,
        jobs: collections.abc.Iterable[job_pb2.Job] | None = ...,
        next_page: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["jobs", b"jobs", "next_page", b"next_page"]) -> None: ...

global___ListJobsResponse = ListJobsResponse

@typing_extensions.final
class JobLog(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    JOB_FIELD_NUMBER: builtins.int
    K8S_INFO_FIELD_NUMBER: builtins.int
    LOG_FIELD_NUMBER: builtins.int
    @property
    def job(self) -> job_pb2.Job: ...
    @property
    def k8s_info(self) -> job_pb2.K8sInfo: ...
    @property
    def log(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    def __init__(
        self,
        *,
        job: job_pb2.Job | None = ...,
        k8s_info: job_pb2.K8sInfo | None = ...,
        log: collections.abc.Iterable[builtins.str] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["job", b"job", "k8s_info", b"k8s_info"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["job", b"job", "k8s_info", b"k8s_info", "log", b"log"]) -> None: ...

global___JobLog = JobLog

@typing_extensions.final
class UpdateRunRequest(google.protobuf.message.Message):
    """To update run for a job, identified by token
    required fields in job are: [token, status, updated_at] (dag is optional)
    message is optional
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TOKEN_FIELD_NUMBER: builtins.int
    JOB_FIELD_NUMBER: builtins.int
    UPDATED_AT_FIELD_NUMBER: builtins.int
    MESSAGE_FIELD_NUMBER: builtins.int
    AUTH_INFO_FIELD_NUMBER: builtins.int
    token: builtins.str
    """token to identify the run"""
    @property
    def job(self) -> job_pb2.Job:
        """set status here"""
    @property
    def updated_at(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """should be time.now()"""
    message: builtins.str
    """optional"""
    @property
    def auth_info(self) -> common_pb2.NBXAuthInfo:
        """Required Auth Info"""
    def __init__(
        self,
        *,
        token: builtins.str = ...,
        job: job_pb2.Job | None = ...,
        updated_at: google.protobuf.timestamp_pb2.Timestamp | None = ...,
        message: builtins.str = ...,
        auth_info: common_pb2.NBXAuthInfo | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["auth_info", b"auth_info", "job", b"job", "updated_at", b"updated_at"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["auth_info", b"auth_info", "job", b"job", "message", b"message", "token", b"token", "updated_at", b"updated_at"]) -> None: ...

global___UpdateRunRequest = UpdateRunRequest



"""
This code is used to manage the authentication of the entire ``nbox`` package. For authentication
it will create a ``.nbx`` in the user's home directory (``~/.nbx``, in case of linux) and store
a file called ``secrets.json``. This folder will also contain more information and files that
are used elsewhere as well ex. files generated when takling to any instance.

Users can chose to set configs at a global level for convenience. Here's a list of configs that can be set:

global.workspace_id = '' # this will set the default workspace id for all commands, users can override this by
  passing the ``--workspace-id`` flag
"""
import os
import json
import requests
import webbrowser
from getpass import getpass
from enum import Enum

import nbox.utils as U
from nbox.utils import join, logger


class ConfigString():
  workspace_id = "config.global.workspace_id"
  workspace_name = "config.global.workspace_name"
  cache = "cache"
  username = "username"
  access_token = "access_token"

  def items():
    return [ConfigString.workspace_id, ConfigString.workspace_name, ConfigString.cache]

class NBXClient:
  def __init__(self, nbx_url = "https://app.nimblebox.ai"):
    """
    Single source for all the secrets file. Can persist across multiple processes.
    """
    os.makedirs(U.env.NBOX_HOME_DIR(), exist_ok=True)
    self.fp = join(U.env.NBOX_HOME_DIR(), "secrets.json")

    access_token = U.env.NBOX_USER_TOKEN("")

    # if this is the first time starting this then get things from the nbx-hq
    if not os.path.exists(self.fp):
      if not access_token:
        logger.info(f"Ensure that you put the email ID you have signed up with!")
        _secrets_url = f"{nbx_url}/secrets"
        logger.info(f"Opening: {_secrets_url}")
        webbrowser.open(_secrets_url)
        access_token = getpass("Access Token: ")
      
      sess = requests.Session()
      sess.headers = {"Authorization": f"Bearer {access_token}"}
      
      # Once we have the access token, we can get the secrets
      _u = f"{nbx_url}/api/v1"
      r = sess.get(_u+"/user/account_details")
      r.raise_for_status()
      try:
        data = r.json()["data"]
        username = data["username"]
        email = data["email"]
      except Exception as e:
        logger.error(f"Could not get the username and email from the response")
        logger.error(f"This should not have happened, please contact NimbleBox support.")
        raise e

      workspace_id = input("Set the default workspace ID (blank means personal): ").strip()
      workspace_details = {"workspace_name": ""}
      
      if workspace_id:
        # fetch some more data about the workspace
        r = sess.get(_u+"/workspace")
        r.raise_for_status()
        try:
          data = r.json()["data"]
          workspace_details = list(filter(lambda x: x["workspace_id"] == workspace_id, data))
          if len(workspace_details) == 0:
            logger.error(f"Could not find the workspace ID: {workspace_id}. Please check the workspace ID and try again.")
            raise Exception("Invalid workspace ID")
          workspace_details = workspace_details[0]
        except Exception as e:
          logger.error(f"Could not get the workspace details from the response")
          logger.error(f"This should not have happened, please contact NimbleBox support.")
          raise e
      else:
        workspace_id = "personal"
      logger.info(f"Setting the default workspace ID to: {workspace_id}")

      # create the objects
      self.secrets = {
        "email": email,
        "access_token": access_token,
        "nbx_url": nbx_url,
        "username": username,

        # config values that can be set by the user for convenience
        ConfigString.workspace_id: workspace_id,
        ConfigString.workspace_name: workspace_details["workspace_name"],

        # now cache the information about this workspace
        ConfigString.cache: {workspace_id: workspace_details},
      }
      # for k,v in self.secrets.items():
      #   print(type(k), k, "::", type(v), v)
      with open(self.fp, "w") as f:
        f.write(repr(self))
      os.makedirs(U.join(U.env.NBOX_HOME_DIR(), ".cache"), exist_ok=True)
      os.makedirs(U.join(U.env.NBOX_HOME_DIR(), "relics"), exist_ok=True)
      logger.info("Successfully created secrets!")
    else:
      with open(self.fp, "r") as f:
        self.secrets = json.load(f)
      logger.debug("Successfully loaded secrets!")

  def __repr__(self):
    return json.dumps(self.secrets, indent=2)

  def get(self, item, default=None, reload: bool = False):
    if reload:
      with open(self.fp, "r") as f:
        self.secrets = json.load(f)
    return self.secrets.get(item, default)

  def put(self, item, value, persist: bool = False):
    self.secrets[item] = value
    if persist:
      with open(self.fp, "w") as f:
        f.write(repr(self))


def init_secret():
  # add any logic here for creating secrets
  if not U.env.NBOX_NO_AUTH(False):
    secret = NBXClient()
    logger.info(f"Current workspace id: {secret.get(ConfigString.workspace_id)} ({secret.get(ConfigString.workspace_name)})")
    return secret
  else:
    logger.info(f"Skipping authentication as NBOX_NO_AUTH is set to True")
  return None

secret = init_secret()



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import collections.abc
import common_pb2
import dag_pb2
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import google.protobuf.timestamp_pb2
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class Job(google.protobuf.message.Message):
    """Job is the definition of a Job ceated by nbox, with a schedule"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _Status:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _StatusEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[Job._Status.ValueType], builtins.type):  # noqa: F821
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        UNKNOWN: Job._Status.ValueType  # 0
        NOT_SET: Job._Status.ValueType  # 1
        """when user creates from FE"""
        PREPARING: Job._Status.ValueType  # 2
        """if the run is being prepared"""
        PAUSED: Job._Status.ValueType  # 3
        """if user from FE/nbox said pause"""
        SCHEDULED: Job._Status.ValueType  # 4
        """is a run is scheduled"""
        ACTIVE: Job._Status.ValueType  # 5
        """run is running"""
        ERROR: Job._Status.ValueType  # 6
        """run errored -> paused"""
        COMPLETED: Job._Status.ValueType  # 7
        """last run done and there is nothing scheduled"""
        ARCHIVED: Job._Status.ValueType  # 8
        """when user deletes this job"""
        CANCELLED: Job._Status.ValueType  # 9
        """run status if job is pasued or archive"""
        KILLED: Job._Status.ValueType  # 10
        """run status if job is archived or paused while run is active"""

    class Status(_Status, metaclass=_StatusEnumTypeWrapper): ...
    UNKNOWN: Job.Status.ValueType  # 0
    NOT_SET: Job.Status.ValueType  # 1
    """when user creates from FE"""
    PREPARING: Job.Status.ValueType  # 2
    """if the run is being prepared"""
    PAUSED: Job.Status.ValueType  # 3
    """if user from FE/nbox said pause"""
    SCHEDULED: Job.Status.ValueType  # 4
    """is a run is scheduled"""
    ACTIVE: Job.Status.ValueType  # 5
    """run is running"""
    ERROR: Job.Status.ValueType  # 6
    """run errored -> paused"""
    COMPLETED: Job.Status.ValueType  # 7
    """last run done and there is nothing scheduled"""
    ARCHIVED: Job.Status.ValueType  # 8
    """when user deletes this job"""
    CANCELLED: Job.Status.ValueType  # 9
    """run status if job is pasued or archive"""
    KILLED: Job.Status.ValueType  # 10
    """run status if job is archived or paused while run is active"""

    @typing_extensions.final
    class FeatureGatesEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.str
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    @typing_extensions.final
    class Schedule(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        START_FIELD_NUMBER: builtins.int
        END_FIELD_NUMBER: builtins.int
        CRON_FIELD_NUMBER: builtins.int
        @property
        def start(self) -> google.protobuf.timestamp_pb2.Timestamp: ...
        @property
        def end(self) -> google.protobuf.timestamp_pb2.Timestamp: ...
        cron: builtins.str
        def __init__(
            self,
            *,
            start: google.protobuf.timestamp_pb2.Timestamp | None = ...,
            end: google.protobuf.timestamp_pb2.Timestamp | None = ...,
            cron: builtins.str = ...,
        ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["end", b"end", "start", b"start"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["cron", b"cron", "end", b"end", "start", b"start"]) -> None: ...

    ID_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    CREATED_AT_FIELD_NUMBER: builtins.int
    CODE_FIELD_NUMBER: builtins.int
    RESOURCE_FIELD_NUMBER: builtins.int
    FEATURE_GATES_FIELD_NUMBER: builtins.int
    OBSOLETE_AUTH_INFO_FIELD_NUMBER: builtins.int
    SCHEDULE_FIELD_NUMBER: builtins.int
    DAG_FIELD_NUMBER: builtins.int
    PAUSED_FIELD_NUMBER: builtins.int
    STATUS_FIELD_NUMBER: builtins.int
    K8SINFO_FIELD_NUMBER: builtins.int
    id: builtins.str
    """unique id of the job"""
    name: builtins.str
    """user defined name of the job
    // user provided description of the job
    string description = 3;
    """
    @property
    def created_at(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """time created"""
    @property
    def code(self) -> common_pb2.Code: ...
    @property
    def resource(self) -> common_pb2.Resource: ...
    @property
    def feature_gates(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]: ...
    @property
    def OBSOLETE_auth_info(self) -> common_pb2.NBXAuthInfo:
        """some blank space for future fields

        reserved 11;
        Deprecated: use auth_info in JobRequest message instead
        """
    @property
    def schedule(self) -> global___Job.Schedule:
        """schedule of the job"""
    @property
    def dag(self) -> dag_pb2.DAG:
        """nbox created DAG information
        get this from dag.proto
        """
    paused: builtins.bool
    """if the job is/has to be paused"""
    status: global___Job.Status.ValueType
    """Job Status, as used within WS"""
    @property
    def k8sInfo(self) -> global___K8sInfo: ...
    def __init__(
        self,
        *,
        id: builtins.str = ...,
        name: builtins.str = ...,
        created_at: google.protobuf.timestamp_pb2.Timestamp | None = ...,
        code: common_pb2.Code | None = ...,
        resource: common_pb2.Resource | None = ...,
        feature_gates: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
        OBSOLETE_auth_info: common_pb2.NBXAuthInfo | None = ...,
        schedule: global___Job.Schedule | None = ...,
        dag: dag_pb2.DAG | None = ...,
        paused: builtins.bool = ...,
        status: global___Job.Status.ValueType = ...,
        k8sInfo: global___K8sInfo | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["OBSOLETE_auth_info", b"OBSOLETE_auth_info", "code", b"code", "created_at", b"created_at", "dag", b"dag", "k8sInfo", b"k8sInfo", "resource", b"resource", "schedule", b"schedule"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["OBSOLETE_auth_info", b"OBSOLETE_auth_info", "code", b"code", "created_at", b"created_at", "dag", b"dag", "feature_gates", b"feature_gates", "id", b"id", "k8sInfo", b"k8sInfo", "name", b"name", "paused", b"paused", "resource", b"resource", "schedule", b"schedule", "status", b"status"]) -> None: ...

global___Job = Job

@typing_extensions.final
class K8sInfo(google.protobuf.message.Message):
    """Some k8s cluster resource info"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CRON_JOB_FIELD_NUMBER: builtins.int
    PODS_FIELD_NUMBER: builtins.int
    cron_job: builtins.str
    """Name of CronJob obj in k8s"""
    @property
    def pods(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """Names of pods in k8s"""
    def __init__(
        self,
        *,
        cron_job: builtins.str = ...,
        pods: collections.abc.Iterable[builtins.str] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["cron_job", b"cron_job", "pods", b"pods"]) -> None: ...

global___K8sInfo = K8sInfo



"""
NimbleBox LMAO is our general purpose observability tool for any kind of computation you might have.
"""

# f"{lmao.project_id}/{lmao.run.experiment_id}/model.pkl"
# drift detection and all
# run.log_dataset(
#     dataset_name='train',
#     features=X_train,
#     predictions=y_pred_train,
#     actuals=y_train,
# )
# run.log_dataset(
#     dataset_name='test',
#     features=X_test,
#     predictions=y_pred_test,
#     actuals=y_test,
# )

import os
import re
import sys
import shlex
import zipfile
from git import Repo
from json import dumps
from functools import lru_cache
from requests import Session
from typing import Dict, Any, List, Optional, Union
from subprocess import Popen
from google.protobuf.field_mask_pb2 import FieldMask

import nbox.utils as U
from nbox import Instance
from nbox.utils import logger, SimplerTimes
from nbox.auth import secret, ConfigString
from nbox.nbxlib.tracer import Tracer
from nbox.relics import RelicsNBX
from nbox.jobs import Job, upload_job_folder
from nbox.hyperloop.common_pb2 import Resource
from nbox.init import nbox_grpc_stub
from nbox.messages import message_to_dict
from nbox.hyperloop.nbox_ws_pb2 import UpdateJobRequest
from nbox.hyperloop.job_pb2 import Job as JobProto
from nbox.hyperloop.common_pb2 import NBXAuthInfo

# all the sublime -> hyperloop stuff
from nbox.sublime.lmao_rpc_client import (
  LMAO_Stub, # main stub class
  Record, File, FileList, AgentDetails, RunLog,
  Run, InitRunRequest, ListProjectsRequest
)
from nbox.observability.system import SystemMetricsLogger


"""
functional components of LMAO
"""

@lru_cache()
def get_lmao_stub() -> LMAO_Stub:
  username = username or secret.get(ConfigString.username)
  workspace_id = workspace_id or secret.get(ConfigString.workspace_id)

  # prepare the URL
  id_or_name = f"monitoring-{workspace_id}"
  logger.info(f"Instance id_or_name: {id_or_name}")
  logger.debug(f"workspace_id: {workspace_id}")
  instance = Instance(id_or_name, workspace_id = workspace_id)
  try:
    open_data = instance.open_data
  except AttributeError:
    raise Exception(f"Is instance '{instance.project_id}' running?")
  build = "build"
  if "app.c." in secret.get("nbx_url"):
    build = "build.c"
  url = f"https://server-{open_data['url']}.{build}.nimblebox.ai/"
  logger.debug(f"URL: {url}")

  # create a session with the auth header
  _session = Session()
  _session.cookies.update(instance.stub_ws_instance._session.cookies)

  # self.lmao = LMAO_Stub(url = "http://127.0.0.1:8080", session = _session) # debug
  lmao_stub = LMAO_Stub(url = url, session = _session)
  return lmao_stub

def get_record(k: str, v: Union[int, float, str]) -> Record:
  """Function to create a Record protobuf object from a key and value."""
  _tv = type(v)
  assert _tv in [int, float, str], f"[key = {k}] '{_tv}' is not a valid type"
  _vt = {
    int: Record.DataType.INTEGER,
    float: Record.DataType.FLOAT,
    str: Record.DataType.STRING,
  }[_tv]
  record = Record(key = k, value_type = _vt)
  if _tv == int:
    record.integer_data.append(v)
  elif _tv == float:
    record.float_data.append(v)
  elif _tv == str:
    record.string_data.append(v)
  return record

def get_git_details(folder):
  """If there is a .git folder in the folder, return some details for that."""
  repo = Repo(folder)

  # check for any unstaged files
  uncommited_files = {}
  diff = repo.index.diff(None)
  for f in diff:
    path = f.a_path or f.b_path # when new file is added, a_path is None
    uncommited_files[path] = f.change_type
  if uncommited_files:
    logger.warning(f"Uncommited files: {uncommited_files}")

  # get the remote url
  try:
    remote_url = repo.remote().url

    # clean the remote_url because it can sometimes contain the repo token as well.
    # this can become security hazard. so if you have an example, that is not suppported:
    #   go ahead, make a PR!
    if "github.com" in remote_url:
      remote_url = re.sub(r"ghp_\w+@", "", remote_url)
  except ValueError:
    remote_url = None

  # get the size of the repository
  size = None
  for line in repo.git.count_objects("-v").splitlines():
    if line.startswith("size:"):
      size = int(line[len("size:") :].strip())
  if size > (1 << 30):
    logger.warning(f"Repository size over 1GB, you might want to work on it")

  return {
    "remote_url": remote_url,
    "branch": repo.active_branch.name,
    "commit": repo.head.commit.hexsha,
    "uncommited_files": uncommited_files,
    "untracked_files": repo.untracked_files,
    "size": size,
  }

"""
Client library that the user will use to interact with the LMAO server.
"""

class _lmaoConfig:
  kv = {}
  def set(
    workspace_id: str = "",
    project_id: Optional[str] = "",
    project_name: Optional[str] = "",
    experiment_id: Optional[str] = "",
    metadata: Dict[str, Any] = {},
    save_to_relic: bool = False,
    enable_system_monitoring: bool = False,
    store_git_details: bool = True,
    args = (),
    kwargs = {},
  ) -> None:
    _lmaoConfig.kv = {
      "workspace_id": workspace_id,
      "project_id": project_id,
      "project_name": project_name,
      "experiment_id": experiment_id,
      "metadata": metadata,
      "save_to_relic": save_to_relic,
      "enable_system_monitoring": enable_system_monitoring,
      "store_git_details": store_git_details,
      "args": args,
      "kwargs": kwargs,
    }

  def clear() -> None:
    _lmaoConfig.kv = {}

  def json() -> str:
    # return dumps({k:v for k,v in _lmaoConfig.kv.items() if k not in ["args", "kwargs"]})
    return dumps(_lmaoConfig.kv)


class Lmao():
  def __init__(
    self,
    project_name: Optional[str] = "",
    project_id: Optional[str] = "",
    experiment_id: Optional[str] = "",
    metadata: Dict[str, Any] = {},
    save_to_relic: bool = False,
    enable_system_monitoring: bool = False,
    store_git_details: bool = True,
    workspace_id: str = "",
  ) -> None:
    """`Lmao` is the client library for using NimbleBox Monitoring. It talks to your monitoring instance running on your build
    and stores the information in the `project_name` or `project_id`. This object inherently doesn't care what you are actually
    logging and rather concerns itself with ensuring storage.

    **Note**: All arguments are optional, if the _lmaoConfig is set.
    
    Args:
      project_name (str, optional): The name of the project. Defaults to "".
      project_id (str, optional): The id of the project. Defaults to "".
      experiment_id (str, optional): The id of the experiment. Defaults to "".
      metadata (Dict[str, Any], optional): Any metadata that you want to store. Defaults to {}.
      save_to_relic (bool, optional): Whether to save the data to the relic. Defaults to False.
      enable_system_monitoring (bool, optional): Whether to enable system monitoring. Defaults to False.
      store_git_details (bool, optional): Whether to store git details. Defaults to True.
      workspace_id (str, optional): The id of the workspace. Defaults to "".
    """

    self.config = _lmaoConfig.kv

    if _lmaoConfig.kv:
      # load all the values from the config
      project_name = _lmaoConfig.kv["project_name"]
      project_id = _lmaoConfig.kv["project_id"]
      experiment_id = _lmaoConfig.kv["experiment_id"]
      metadata = _lmaoConfig.kv["metadata"]
      save_to_relic = _lmaoConfig.kv["save_to_relic"]
      enable_system_monitoring = _lmaoConfig.kv["enable_system_monitoring"]
      store_git_details = _lmaoConfig.kv["store_git_details"]
      workspace_id = _lmaoConfig.kv["workspace_id"]

    self.project_name = project_name
    self.project_id = project_id
    self.experiment_id = experiment_id
    self.metadata = metadata
    self.save_to_relic = save_to_relic
    self.enable_system_monitoring = enable_system_monitoring
    self.store_git_details = store_git_details
    self.workspace_id = workspace_id or secret.get(ConfigString.workspace_id)
    
    # now set the supporting keys
    self.nbx_job_folder = U.env.NBOX_JOB_FOLDER("")
    self._total_logged_elements = 0 # this variable keeps track for logging
    self.completed = False
    self.tracer = None
    self.username = secret.get("username")
    self.relic: RelicsNBX = None
    self.system_monitoring: SystemMetricsLogger = None
    self._nbx_run_id = None
    self._nbx_job_id = None

    # create connection and initialise the run
    self._init(project_name, project_id, config = metadata)

  def __del__(self):
    if self.system_monitoring is not None:
      self.system_monitoring.stop()

  def _get_name_id(self, project_id: str = None, project_name: str = None):
    matching_projects = self.lmao.list_projects(
      ListProjectsRequest(
        workspace_id = self.workspace_id,
        project_id_or_name = project_id if project_id else project_name,
      )
    )

    # case for one matching project
    if len(matching_projects.projects) == 1:
      project_name = matching_projects.projects[0].project_name
      project_id = matching_projects.projects[0].project_id
    elif len(matching_projects.projects) > 1:
      if project_id:
        logger.error(f"Multiple entries for {project_id} found, something went wrong from our side.")
        raise Exception(f"Database duplicate entry for project_id: {project_id}")
      else:
        logger.error(f"Project '{project_name}' found multiple times, please use project_id instead.")
        raise Exception("Ambiguous project name, please use project_id instead.")
    else:
      # case for no matching projects returned
      # means we have to initialise this project first and then we will pull the details after the init_run
      if project_id:
        raise Exception(f"Project with id {project_id} not found, please create a new one with project_name")
      logger.info(f"Project '{project_name}' not found, will create a new one.")

    return project_id, project_name

  def _init(self, project_name, project_id, config: Dict[str, Any] = {}):
    # create a tracer object that will load all the information
    tracer = Tracer(start_heartbeat = False)
    self._nbx_run_id = tracer.run_id
    self._nbx_job_id = tracer.job_id
    del tracer
    self.lmao = get_lmao_stub()

    # do a quick lookup and see if the project exists, if not, create it
    if self.config:
      project_id = self.config["project_id"]
      project_name = self.config["project_name"]
    else:
      if project_id:
        project_id, project_name = self._get_name_id(project_id=project_id)
        if not project_id:
          raise Exception(f"Project with id {project_id} not found, please create a new one with project_name")
      elif project_name:
        project_id, project_name = self._get_name_id(project_name=project_name)
        if not project_id:
          logger.info(f"Project '{project_name}' not found, create one from dashboard.")
      else:
        raise Exception("provide either `project_id` or `project_name`")

    # this is the config value that is used to store data on the plaform, user cannot be allowed to have
    # like a full access to config values
    log_config: Dict[str, Any] = {
      "user_config": config
    }

    # check if the current folder from where this code is being executed has a .git folder
    # NOTE: in case of NBX-Jobs the current folder ("./") is expected to contain git by default
    log_config["git"] = None
    if os.path.exists(".git") and self.store_git_details:
      log_config["git"] = get_git_details("./")

    # continue as before
    self._agent_details = AgentDetails(
      type = AgentDetails.NBX.JOB,
      nbx_job_id = self._nbx_job_id or "",
      nbx_run_id = self._nbx_run_id or "",
    )

    if self.experiment_id:
      run_details = self.lmao.get_run_details(Run(experiment_id = self.experiment_id, project_id=project_id))
      if not run_details:
        # TODO: Make a custom exception of this
        raise Exception("Server Side exception has occurred, Check the log for details")
      if run_details.experiment_id:
        # means that this run already exists so we need to make an update call
        ack = self.lmao.update_run_status(Run(
          project_id = project_id,
          experiment_id = run_details.experiment_id,
          agent = self._agent_details,
          update_keys = ["agent"],
        ))
        if not ack.success:
          raise Exception(f"Failed to update run status! {ack.message}")
    else:
      run_details = self.lmao.init_run(
        InitRunRequest(
          agent_details=self._agent_details,
          created_at = SimplerTimes.get_now_i64(),
          project_name = project_name,
          project_id = project_id,
          config = dumps(log_config),
        )
      )

    self.project_name = project_name
    self.project_id = run_details.project_id
    self.metadata = config

    self.run = run_details
    logger.info(f"Created a new LMAO run")
    logger.info(f" project: {self.project_name} ({self.project_id})")
    logger.info(f"      id: {self.run.experiment_id}")
    logger.info(f"    link: https://app.nimblebox.ai/workspace/{self.workspace_id}/monitoring/{self.project_id}/{self.run.experiment_id}")

    # now initialize the relic
    if self.save_to_relic:
      # The relic will be the project id
      self.relic = RelicsNBX(LMAO_RELIC_NAME, self.workspace_id, create = True)
      logger.info(f"Will store everything in relic '{LMAO_RELIC_NAME}' folder: {self.experiment_prefix}")

    # system metrics monitoring, by default is enabled optionally turn it off
    if self.enable_system_monitoring:
      self.system_monitoring = SystemMetricsLogger(self)
      self.system_monitoring.start()

  @property
  def experiment_prefix(self):
    return f"{self.project_name}/{self.run.experiment_id}/"

  """The functions below are the ones supposed to be used."""

  @lru_cache(maxsize=1)
  def get_relic(self):
    """Get the underlying Relic for more advanced usage patterns."""
    return RelicsNBX(LMAO_RELIC_NAME, self.workspace_id, create = True, prefix = f"{self.project_name}/{self.run.experiment_id}")

  def log(self, y: Dict[str, Union[int, float, str]], step = None, *, log_type: str = RunLog.LogType.USER):
    """Log a single level dictionary to the platform at any given step. This function does not really care about the
    information that is being logged, it just logs it to the platform."""
    if self.completed:
      raise Exception("Run already completed, cannot log more data!")

    step = step if step is not None else SimplerTimes.get_now_i64()
    if step < 0:
      raise Exception("Step must be <= 0")
    run_log = RunLog(experiment_id = self.run.experiment_id, project_id=self.project_id, log_type=log_type)
    for k,v in y.items():
      # TODO:@yashbonde replace Record with RecordColumn
      record = get_record(k, v)
      record.step = step
      run_log.data.append(record)

    ack = self.lmao.on_log(run_log)
    if not ack.success:
      logger.error(f"  >> Server Error\n{ack.message}")
      raise Exception("Server Error")

    self._total_logged_elements += 1

  def save_file(self, *files: List[str]):
    """
    Register a file save. User should be aware of some structures that we follow for standardizing the data.
    All the experiments are going to be tracked under the following pattern:

    #. `relic_name` is going to be the experiment ID, so any changes to the name will not affect relic storage
    #. `{experiment_id}(_{job_id}@{experiment_id})` is the name of the folder which contains all the artifacts in the experiment.

    If relics is not enabled, this function will simply log to the LMAO DB.

    dk.save_file("foo.t", "/bar/", "baz.t", "/bar/roo/")
    """
    logger.info(f"Saving files: {files}")

    # manage all the complexity of getting the list of RelicFile
    all_files = []
    for folder_or_file in files:
      if os.path.isfile(folder_or_file):
        all_files.append(folder_or_file)
      elif os.path.isdir(folder_or_file):
        all_files.extend(U.get_files_in_folder(folder_or_file, self.username, self.workspace_id))
      else:
        raise Exception(f"File or Folder not found: {folder_or_file}")

    logger.debug(f"Storing {len(all_files)} files")
    if self.save_to_relic:
      relic = self.get_relic()
      logger.info(f"Uploading files to relic: {relic}")
      for f in all_files:
        relic.put(f)

    # TODO: @yashbonde log the files in the LMAO DB for sanity, currently this is a no-op, keeping it here so one day
    # when we add something cool we can use this
    # fl = FileList(experiment_id = self.run.experiment_id)
    # fl.files.extend([File(relic_file = RelicFile(name = x)) for x in all_files])
    # self.lmao.on_save(fl)

  def end(self):
    """End the run to declare it complete. This is more of a convinience function than anything else. For example when you
    are monitoring a live API you may never know when the experiment is complete. This function locks the experiment name
    so it can't be modified."""
    if self.completed:
      logger.error("Run already completed, cannot end it again!")
      return None

    logger.info("Ending run")
    ack = self.lmao.on_train_end(self.run)
    if not ack.success:
      logger.error("  >> Server Error")
      for l in ack.message.splitlines():
        logger.error("  " + l)
      raise Exception("Server Error")
    self.completed = True
    if self.enable_system_monitoring:
      self.system_monitoring.stop()

"""
For some Experiences you want to have CLI control so this class manages that

lmao upload fp:fn project_name_or_id
lmao trigger project_name_or_id
lmao open project_name_or_id
"""

@lru_cache()
def get_project_name_id(project_name_or_id: str, workspace_id: str):
  lmao_stub = get_lmao_stub()
  out = lmao_stub.list_projects(ListProjectsRequest(workspace_id=workspace_id, project_id_or_name=project_name_or_id))
  if not out.projects:
    logger.error(f"Project: {project_name_or_id} not found")
    logger.error(f"Will automatically create one")
    name = project_name_or_id
    id = None
  elif len(out.projects) > 1:
    raise ValueError(f"Multiple projects found for: {project_name_or_id}")
  else:
    p = out.projects[0]
    name = p.project_name
    id = p.project_id
  return name, id


class LmaoCLI:
  # This class will only manage the things that need to talk to the DB and everything else is offloaded
  def upload(
    self,
    init_path: str,
    project_name_or_id: str,
    workspace_id: str = "",
    trigger: bool = False,

    # all the arguments for the git thing
    untracked: bool = False,
    untracked_no_limit: bool = False,

    # all the things for resources
    resource_cpu: str = "100m",
    resource_memory: str = "128Mi",
    resource_disk_size: str = "1Gi",
    resource_gpu: str = "none",
    resource_gpu_count: str = "0",
    resource_timeout: int = 120_000,
    resource_max_retries: int = 2,

    # the following arguments are used for the initialisation of lmao class
    save_to_relic: bool = True,
    enable_system_monitoring: bool = False,

    # the following things are needed for the different modules in the process
    relics_kwargs: Dict[str, Any] = {},

    # any other things to pass to the function / class being called
    **run_kwargs
  ):
    """Upload and register a new run for a NBX-LMAO project.

    Args:
      init_path (str): This can be a path to a `folder` or can be optionally of the structure `fp:fn` where `fp`
        is the path to the file and `fn` is the function name.
      project_name_or_id (str): The name or id of the LMAO project.
      workspace_id (str, optional): If nor provided, defaults to global config
      trigger (bool, optional): Defaults to False. If True, will trigger the run after uploading.
      untracked (bool, optional): If True, then untracked files below 1MB will also be zipped and uploaded. Defaults to False.
      untracked_no_limit (bool, optional): If True, then all untracked files will also be zipped and uploaded. Defaults to False.
      resource_cpu (str, optional): Defaults to "100m". The CPU resource to allocate to the run.
      resource_memory (str, optional): Defaults to "128Mi". The memory resource to allocate to the run.
      resource_disk_size (str, optional): Defaults to "1Gi". The disk size resource to allocate to the run.
      resource_gpu (str, optional): Defaults to "none". The GPU resource to use.
      resource_gpu_count (str, optional): Defaults to "0". Number of GPUs allocated to the experiment.
      resource_timeout (int, optional): Defaults to 120_000. The timeout between two consecutive runs, honoured but not guaranteed.
      resource_max_retries (int, optional): Defaults to 2. The maximum number of retries for a run.
      **run_kwargs: These are the kwargs that will be passed to your Operator.
    """
    # reconstruct the entire CLI command so we can show it in the UI
    workspace_id = workspace_id or secret.get(ConfigString.workspace_id)
    reconstructed_cli_comm = (
      f"nbx lmao upload '{init_path}' '{project_name_or_id}'"
      f" --workspace_id '{workspace_id}'"
      f" --resource_cpu '{resource_cpu}'"
      f" --resource_memory '{resource_memory}'"
      f" --resource_disk_size '{resource_disk_size}'"
      f" --resource_gpu '{resource_gpu}'"
      f" --resource_gpu_count '{resource_gpu_count}'"
      f" --resource_timeout {resource_timeout}"
      f" --resource_max_retries {resource_max_retries}"
    )
    if trigger:
      reconstructed_cli_comm += " --trigger"
    if untracked:
      reconstructed_cli_comm += " --untracked"
    if untracked_no_limit:
      reconstructed_cli_comm += " --untracked_no_limit"
    if save_to_relic:
      reconstructed_cli_comm += " --save_to_relic"
    if enable_system_monitoring:
      reconstructed_cli_comm += " --enable_system_monitoring"

    if resource_max_retries < 1:
      logger.error(f"max_retries must be >= 1. Got: {resource_max_retries}\n  Fix: set --max_retries=2")
      raise ValueError()

    # uncommenting these since it can contain sensitive information, kept just in case for taking dictionaries as input
    # if relics_kwargs:
    #   rks = str(relics_kwargs).replace("'", "")
    #   reconstructed_cli_comm += f" --relics_kwargs '{rks}'"
    
    for k, v in run_kwargs.items():
      if type(v) == bool:
        reconstructed_cli_comm += f" --{k}"
      else:
        reconstructed_cli_comm += f" --{k} '{v}'"
    logger.debug(f"command: {reconstructed_cli_comm}")

    # clean user args
    resource_gpu_count = str(resource_gpu_count)
    if untracked_no_limit and not untracked:
      logger.debug("untracked_no_limit is True but untracked is False. Setting untracked to True")
      untracked = True

    # first step is to get all the relevant information from the DB
    workspace_id = workspace_id or secret.get(ConfigString.workspace_id)
    project_name, project_id = get_project_name_id(project_name_or_id, workspace_id)
    logger.info(f"Project: {project_name} ({project_id})")

    # create a call init run and get the experiment metadata
    init_folder, _ = os.path.split(init_path)
    init_folder = init_folder or "."
    if os.path.exists(U.join(init_folder, ".git")):
      git_det = get_git_details(init_folder)
    else:
      git_det = {}


    job_name = "nbxj_" + project_name[:15]
    job = Job(job_name = job_name, workspace_id = workspace_id)
    lmao_stub = get_lmao_stub()

    _metadata = {
      "user_config": run_kwargs,
      "git": git_det,
      "resource": message_to_dict(Resource(
        cpu = resource_cpu,
        memory = resource_memory,
        disk_size = resource_disk_size,
        gpu = resource_gpu,
        gpu_count = resource_gpu_count,
        timeout = resource_timeout,
        max_retries = resource_max_retries,
      )),
      "cli": reconstructed_cli_comm,
      "lmao": {
        "save_to_relic": save_to_relic,
        "enable_system_monitoring": enable_system_monitoring,
      }
    }

    run = lmao_stub.init_run(InitRunRequest(
      agent_details = AgentDetails(
        type = AgentDetails.NBX.JOB,
        nbx_job_id = job.id, # run id will have to be updated from the job
      ),
      created_at = SimplerTimes.get_now_i64(),
      config = dumps(_metadata),
      project_id = project_id,
      project_name = project_name,
    ))
    if not project_id:
      logger.info(f"Project {project_name} created with id: {run.project_id}")
    logger.info(f"Run ID: {run.experiment_id}")

    # connect to the relic
    r_keys = set(relics_kwargs.keys())
    valid_keys = {"bucket_name", "region", "nbx_resource_id", "nbx_integration_token"}
    extra_keys = r_keys - valid_keys
    if extra_keys:
      logger.error("Unknown arguments found:\n  * " + "\n  * ".join(extra_keys))
      raise RuntimeError("Unknown arguments found in the Relic")
    relic = RelicsNBX(LMAO_RELIC_NAME, workspace_id = workspace_id, create = True, **relics_kwargs)

    # create a git patch and upload it to relics
    if git_det:
      _zf = U.join(init_folder, "untracked.zip")
      zip_file = zipfile.ZipFile(_zf, "w", zipfile.ZIP_DEFLATED)
      patch_file = U.join(init_folder, "nbx_auto_patch.diff")
      f = open(patch_file, "w")
      Popen(shlex.split(f"git diff {' '.join(git_det['uncommited_files'])}"), stdout=f, stderr=sys.stderr).wait()
      f.close()
      zip_file.write(patch_file, arcname = patch_file)
      if untracked:
        untracked_files = git_det["untracked_files"] # what to do with these?
        if untracked_files:
          # see if any file is larger than 10MB and if so, warn the user
          warn_once = False
          for f in untracked_files:
            if os.path.getsize(f) > 1e7 and not untracked_no_limit:
              logger.warning(f"File: {f} is larger than 10MB and will not be available in sync")
              logger.warning("  Fix: use git to track small files, avoid large files")
              logger.warning("  Fix: nbox.Relics can be used to store large files")
              logger.warning("  Fix: use --untracked-no-limit to upload all files")
              warn_once = True
              continue
            zip_file.write(f, arcname = f)
      relic.put_to(_zf, f"{project_name}/{run.experiment_id}/git.zip")
      os.remove(_zf)
      os.remove(patch_file)

    # tell the server that this run is being scheduled so atleast the information is visible on the dashboard
    upload_job_folder(
      "job",
      init_folder = init_path,
      name = "nbxj_" + project_name[:15], # keep only 20 chars

      # pass along the resource requirements
      resource_cpu = resource_cpu,
      resource_memory = resource_memory,
      resource_disk_size = resource_disk_size,
      resource_gpu = resource_gpu,
      resource_gpu_count = resource_gpu_count,
      resource_timeout = resource_timeout,
      resource_max_retries = resource_max_retries,
    )

    if trigger:
      logger.debug(f"Running job '{job.name}' ({job.id})")

      # create the serialisable config
      _lmaoConfig.clear()
      _lmaoConfig.set(
        workspace_id = workspace_id,
        project_name = run.project_name,
        project_id = run.project_id,
        experiment_id = run.experiment_id,
        metadata = _metadata,
        save_to_relic = save_to_relic,
        enable_system_monitoring = enable_system_monitoring,
        args = (),
        kwargs = run_kwargs,
      )
      
      # from pprint import pprint
      # pprint(_lmaoConfig.kv)
      # exit()

      # put the items in the relic and create a tag from it
      fp = f"{project_name}/{run.experiment_id}"
      relic.put_object(fp+"/init.pkl", _lmaoConfig.kv)
      tag = f"{LMAO_JOB_TYPE_PREFIX}-{fp}"
      logger.info(f"Run tag: {tag}")
      job.trigger(tag)

      # clear so the rest of the program doesn't get affected
      _lmaoConfig.clear()

    # finally print the location of the run where the users can track this
    logger.info(f"Run location: https://app.nimblebox.ai/workspace/{workspace_id}/monitoring/{project_id}/{run.experiment_id}")

# do not change these it can become a huge pain later on
LMAO_RELIC_NAME = "experiments"
LMAO_JOB_TYPE_PREFIX = "NBXLmao"
LMAO_ENV_VAR_PREFIX = "LMAO_"



"""
## Initialization

This file loads first and is responsible for setting up all the global networking items.

Build systems are an extremely important detail of any software project. When they work,
they can feel like magic: you execute a command, and after a series of potentially
complicated steps, a working binary (or other artifact) is produced! When they don't work,
they can feel like confusing, non-transparent roadblocks that you wish didn't exist. This
is typical for any powerful tool: magic or a headache depending on the day and the task.
"""

import os
import json
import grpc
import requests

try:
  from packaging.version import parse
except ImportError:
  from pip._vendor.packaging.version import parse


from nbox.auth import secret, ConfigString
from nbox.utils import logger, env
from nbox.subway import Sub30
from nbox.hyperloop.nbox_ws_pb2_grpc import WSJobServiceStub
from nbox.hyperloop.serve_pb2_grpc import ServingServiceStub, ModelServiceStub
from nbox.version import __version__


def __create_channel(channel_name) -> grpc.Channel:
  """Create a gRPC channel with the Webserver, this will return `webserver_channel` with credentials and ssl.
  
  Args:
    channel_name (str): Name of the channel, used for logging.
  
  Returns:
    grpc.Channel: A gRPC channel with credentials and ssl.
  """
  token_cred = grpc.access_token_call_credentials(secret.get("access_token"))
  ssl_creds = grpc.ssl_channel_credentials()
  creds = grpc.composite_channel_credentials(ssl_creds, token_cred)
  channel = grpc.secure_channel(secret.get("nbx_url").replace("https://", "dns:/") + ":443", creds)
  future = grpc.channel_ready_future(channel)
  future.add_done_callback(lambda _: logger.debug(f"NBX '{channel_name}' gRPC stub is ready!"))

  return channel

def get_job_stub() -> WSJobServiceStub:
  """Create a gRPC stub with the NBX Webserver, this will initialise `nbox_grpc_stub`
  object which is globally accesible as `nbox.nbox_grpc_stub`. If you find yourself
  using this function, you might want to reconsider your design.
  
  Returns:
    WSJobServiceStub: A gRPC stub with the NBX Webserver.
  """
  channel = __create_channel("WSJobServiceStub")
  stub = WSJobServiceStub(channel)
  logger.debug(f"Connected using stub: {stub.__class__.__name__}")
  return stub


def get_serving_stub() -> ServingServiceStub:
  """Create a gRPC stub with the NBX Serving, this will initialise `nbox_serving_grpc_stub`
  object which is globally accesible as `nbox.nbox_serving_grpc_stub`. If you find yourself
  using this function, you might want to reconsider your design.
  
  Returns:
    ServingServiceStub: A gRPC stub with the NBX Serving.
  """
  channel = __create_channel("ServingServiceStub")
  stub = ServingServiceStub(channel)
  logger.debug(f"Connected using stub: {stub.__class__.__name__}")
  return stub


def get_model_stub() -> ModelServiceStub:
  """Create a gRPC stub with the NBX Serving, this will initialise `nbox_model_grpc_stub`
  object which is globally accesible as `nbox.nbox_model_grpc_stub`. If you find yourself
  using this function, you might want to reconsider your design.

  Returns:
    ModelServiceStub: A gRPC stub with the NBX Serving.
  """
  channel = __create_channel("ModelServiceStub")
  stub = ModelServiceStub(channel)
  logger.debug(f"Connected using stub: {stub.__class__.__name__}")
  return stub


def create_webserver_subway(version: str = "v1", session: requests.Session = None) -> Sub30:
  """Create a Subway object for the NBX Webserver. This is a wrapper around the
  OpenAPI spec plublished by NBX Webserver. It loads the JSON object in `Sub30`
  which allows accesing REST APIs with python "." (dot) notation. If you find yourself
  using this function, you might want to reconsider your design.

  Args:
    version (str, optional): Version of the API to use. Defaults to "v1".
    session (requests.Session, optional): Session to use for the API calls. Defaults to None.

  Returns:
    Sub30: A Subway object for the NBX Webserver.
  """
  _version_specific_url = secret.get("nbx_url") + f"/api/{version}"
  session = session if session != None else nbox_session  # select correct session
  r = session.get(_version_specific_url + "/openapi.json")
  try:
    r.raise_for_status()
  except Exception as e:
    logger.error(f"Could not connect to webserver at {secret.get('nbx_url')}")
    logger.error(e)
    return None

  spec = r.json()
  # secret.put("openapi_spec", spec, True)
  # with open(os.path.join(env.NBOX_HOME_DIR(), "openapi.json"), "w") as f:
  #   f.write(json.dumps(spec, indent=2))
  out = Sub30(_version_specific_url, spec, session)
  logger.debug(f"Connected to webserver at {out}")
  return out


def nbox_version_update():
  """Function to check if the current version of nbox is the latest version. If not, it will
  print a warning message to the user. This function is called on import of the nbox module."""
  # https://stackoverflow.com/questions/28774852/pypi-api-how-to-get-stable-package-version
  def get_version(package):
    """Return version of package on pypi.python.org using json."""
    URL_PATTERN = "https://pypi.python.org/pypi/{package}/json"
    req = requests.get(URL_PATTERN.format(package=package))
    version = parse("0")
    if req.status_code == requests.codes.ok:
      j = json.loads(req.text.encode(req.encoding))
      releases = j.get("releases", [])
      for release in releases:
        ver = parse(release)
        if not ver.is_prerelease:
          version = max(version, ver)
    return version

  latest_version = get_version("nbox")
  if latest_version > parse(__version__):
    logger.warning(
      f"Your version of nbox ({__version__}) is outdated. Some functionalities might not work.\n"
      f"  Fix: Please update to {latest_version} using pip3 install nbox --upgrade\n"
      f"  Fix: Update all the relevant requirements.txt files with nbox[serving]=={latest_version}"
    )


# perform all the initialisations based on the environment variables
if env.NBOX_NO_LOAD_GRPC():
  nbox_grpc_stub = None
  nbox_serving_service_stub = None
  nbox_model_service_stub = None
else:
  nbox_grpc_stub: WSJobServiceStub = get_job_stub()
  nbox_serving_service_stub: ServingServiceStub = get_serving_stub()
  nbox_model_service_stub: ModelServiceStub = get_model_stub()

if env.NBOX_NO_LOAD_WS():
  nbox_session = None
  nbox_ws_v1 = None
else:
  nbox_session = requests.Session()
  nbox_session.headers.update({"Authorization": f"Bearer {secret.get('access_token')}"})
  nbox_ws_v1: Sub30 = create_webserver_subway(version="v1", session=nbox_session)

if not env.NBOX_NO_CHECK_VERSION():
  nbox_version_update()



"""
This is CLI for `nbox`, it is meant to be as simple to use as possible. We using `python-fire` for building
our CLI, which means you can access underlying system with same commands as you use with python.

# SSH into Instances

You can SSH into your instance with the `nbx tunnel` command. This command will create a tunnel
to your instance and start an SSH session.

```bash
nbx tunnel 8000 -i "instance-name"
```
"""

import os
import fire
import jinja2
import tempfile
import webbrowser
from json import dumps
from typing import Dict, Any

import nbox.utils as U
from nbox.jobs import Job, Serve
from nbox.init import nbox_ws_v1
from nbox.auth import init_secret, ConfigString, secret
from nbox.instance import Instance
from nbox.sub_utils.ssh import tunnel
from nbox.relics import RelicsNBX
from nbox.lmao import LmaoCLI
from nbox.version import __version__ as V
from nbox.nbxlib.fire import NBXFire

logger = U.get_logger()


class Config(object):
  def update(self, workspace_id: str):
    """Set global config for `nbox`"""
    secret = init_secret()
    data = secret.get(ConfigString.cache)
    redo = not data or (workspace_id not in data)
    if redo:
      workspaces = nbox_ws_v1.workspace()
      workspace_details = list(filter(lambda x: x["workspace_id"] == workspace_id, workspaces))
      if len(workspace_details) == 0:
        logger.error(f"Could not find the workspace ID: {workspace_id}. Please check the workspace ID and try again.")
        raise Exception("Invalid workspace ID")
      workspace_details = workspace_details[0]
      workspace_name = workspace_details["workspace_name"]
      secret.secrets[ConfigString.cache].update({
        workspace_id: {
          "workspace_id": workspace_id,
          "workspace_name": workspace_name
        }
      })
    else:
      data = data[workspace_id]
      workspace_name = data["workspace_name"]

    secret.put(ConfigString.workspace_id, workspace_id, True)
    secret.put(ConfigString.workspace_name, workspace_name, True)
    logger.info(f"Global Workspace: {workspace_name}")

  def show(self):
    """Pretty print global config for `nbox`"""
    logger.info(
      "\nnbox config:\n" \
      f"  workspace_name: {secret.get(ConfigString.workspace_name)}\n" \
      f"    workspace_id: {secret.get(ConfigString.workspace_id)}\n" \
      f"        username: {secret.get(ConfigString.username)}\n" \
      f"    nbox version: {V}\n" \
      f"             URL: {secret.get('nbx_url')}"
    )

def open_home():
  """Open current NBX platform"""
  from .auth import secret
  import webbrowser
  webbrowser.open(secret.get("nbx_url"))


def get(api_end: str, no_pp: bool = False, **kwargs):
  """GET any API.

  Args:
    api_end (str): something like '/v1/api'
    no_pp (bool, optional): Pretty print. Defaults to False.
  """
  api_end = api_end.strip("/")
  if nbox_ws_v1 == None:
    raise RuntimeError("Not connected to NimbleBox.ai webserver")
  out = nbox_ws_v1
  for k in api_end.split("/"):
    out = getattr(out, k)
  res = out(_method = "get", **kwargs)
  if not no_pp:
    logger.info(dumps(res, indent=2))
  else:
    return res

def login():
  """Re authenticate `nbox`. NOTE: Will remove all added keys to `secrets`"""
  fp = U.join(U.env.NBOX_HOME_DIR(), "secrets.json")
  os.remove(fp)
  init_secret()

def version():
  logger.info("NimbleBox.ai Client Library")
  logger.info(f"    nbox version: {V}")


def why():
  print("\nWhy we build NimbleBox?\n")
  print("  * Artificial intelligence will be the most important technology of the 21st century.")
  print("  * Every major piece of software ever written will need to be upgraded and rewritten.")
  print("  * Energy spent per code token will increase exponentially to handle the bandwidth of AI.")
  print("  * AI is still software and software engineering is hard.")
  print("  * Nimblebox is a general purpose tool to build and manage such operations.")
  print("\nIf you like what we are building, come work with us.\n\nWith Love,\nNimbleBox.ai\n")


class NBXWS_CLI(object):
  def get(self, api: str, H: Dict[str, str] = {}):
    """make a GET call to any `api`, it can either contain '/' or '.'"""
    if nbox_ws_v1 == None:
      raise RuntimeError("Not connected to NimbleBox.ai webserver")
    api = api.strip("/")
    splitter = '.' if '.' in api else '/'
    nbox_ws_v1._session.headers.update(H)
    out = nbox_ws_v1
    for k in api.split(splitter):
      out = getattr(out, k)
    res = out(_method = "get")
    return res

  def post(self, api: str, d: Dict[str, Any] = {}, H: Dict[str, str] = {}):
    """make a POST call to any `api`, it can either contain '/' or '.'"""
    if nbox_ws_v1 == None:
      raise RuntimeError("Not connected to NimbleBox.ai webserver")
    api = api.strip("/")
    splitter = '.' if '.' in api else '/'
    nbox_ws_v1._session.headers.update(H)
    out = nbox_ws_v1
    for k in api.split(splitter):
      out = getattr(out, k)
    res = out(_method = "post", data = d)
    return res

  # def rapidoc(self):
  #   openapi = secret.get("openapi_spec", None)
  #   if openapi == None:
  #     raise RuntimeError("Not connected to NimbleBox.ai webserver")
  #   fp = U.join(U.folder(__file__), "assets", "rapidoc.html")
  #   with open(fp, "r") as f:
  #     temp = jinja2.Template(f.read())
  #   with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix=".html") as f:
  #     f.write(temp.render(openapi = dumps(openapi)))
  #     print(f.name)
  #     webbrowser.open(f"file://{f.name}")


def main():
  component = {
    "build"   : Instance,
    "config"  : Config,
    "get"     : get,
    "jobs"    : Job,
    "lmao"    : LmaoCLI,
    "login"   : login,
    "open"    : open_home,
    "relics"  : RelicsNBX,
    "serve"   : Serve,
    "tunnel"  : tunnel,
    "version" : version,
    "why"     : why,
    "ws"      : NBXWS_CLI,
  }

  fire.Fire(component)
  # NBXFire(component)

if __name__ == "__main__":
  main()



import sys

class TC:
  # terminal colors from https://stackoverflow.com/a/287944/1123955
  HEADER = '\033[95m'
  OKBLUE = '\033[94m'
  OKCYAN = '\033[96m'
  OKGREEN = '\033[92m'
  WARNING = '\033[93m'
  FAIL = '\033[91m'
  ENDC = '\033[0m'
  BOLD = '\033[1m'
  UNDERLINE = '\033[4m'


def colored(st, color):
  return f"\u001b[{30+['black', 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white'].index(color)}m{st}\u001b[0m"

class NBXFire:
  """This is the CLI function for bespoke designed for nbox. Names after the legendary `python-fire` command which served
  us well for many years before we ended up here.
  
  - can load any kind of python object
  - '-' can be as many as you want they would be ignored
  - booleans are parsed as `--flag`
  - strings/int are parsed as `value`
  - /dict/list are parsed as `'["0", 123]'` or `"{a: 1, b: 2}"`
  """
  def help_for_comp(self, comp):
    lines = [
      f'{TC.FAIL}ERROR:{TC.ENDC} Cannot process empty command',
      f'{TC.OKGREEN}USAGE:{TC.ENDC} nbx COMMAND\n',
      'Command can be any of the following:\n',
      '\n'.join([f'  {c}' for c in dir(comp) if not c.startswith('_')]),
    ]
    self._print_and_exit(lines)

  def __init__(self, component):
    # component is how we are representing the object we are working with
    
    # get args and if nothing is passed then run the one time help which is outside the main loop
    args = sys.argv[1:]
    if len(args) == 0:
      lines = [
        f'{TC.FAIL}ERROR:{TC.ENDC} Cannot process empty command',
        f'{TC.OKGREEN}USAGE:{TC.ENDC} nbx COMMAND\n',
        'Command can be any of the following:\n',
        '\n'.join([f'  {c}' for c in dir(component) if not c.startswith('_')]),
      ]
      self._print_and_exit(lines)

    print(args)

    # # this is the args kwargs that will be refreshed every time we go inside a component
    # kwargs = {}
    # args_list = []
    # curr_ptr = 0
    # next_ptr = 0
    # while curr_ptr < len(args):
    #   arg = args[curr_ptr]
    #   if arg.startswith('-'):
    #     arg = arg.lstrip('-')      # remove all the dashes
    #     if arg in dir(component):  # if the arg is a method of the component
          

    
    
    # all_args = []

    # # what will be passed to the function, defined up here in case of partials
    # for i, arg in enumerate(args[1:]):
    #   # we need to do a bunch of analysis of the comp here and then based on those we need to create the required
    #   # data for the different templates
    #   # ex: if the comp is a partial function then we need to get functions details from the underlying function
    #   if type(comp) == partial:
    #     args_list.extend(tuple(comp.args))
    #     kwargs.update(comp.keywords)
    #     comp = comp.func

    #   if arg == '--help':
    #     usage = comp.__doc__
    #     # and this is a class then load the docstring from the __init__
    #     if usage is None and type(comp) == type:
    #       usage = comp.__init__.__doc__

    #     # make lines and exit
    #     lines = [
    #       f"{TC.BOLD}{TC.OKGREEN}Usage:{TC.ENDC} {os.path.basename(sys.argv[0])} {service_level}\n",
    #       f"{TC.BOLD}{TC.OKGREEN}Description:{TC.ENDC} {usage}\n",
    #       f"{TC.BOLD}{TC.OKGREEN}Options:{TC.ENDC}"
    #     ] + [f"  {x}" for x in dir(comp) if not x.startswith('_')]
    #     self._print_and_exit(lines)
      
    #   # get the comp and build the kwargs dict
    #   _comp = getattr(comp, arg, None)
    #   if _comp is None:
    #     # this means the user is probably trying to call this python object so we just break here
    #     all_args = args[i+1:]
    #     break

    #   # overwrite the comp
    #   comp = _comp

    # # by now we have the comp all the things required to run the comp, however they are not in the right order
    # # or all the keys provided are useful or not. here's a list of valid keys
    # # nbx jobs upload 'xyz' -id 'abc def' --bool -u '....'
    # # this is a tricky piece of code to write, so we will use a simple while loop with O(1) just like how
    # # language parsers work.
    # next_arg = ""
    # arg = all_args.pop(0)
    # while arg:
    #   if len(all_args):
    #     next_arg = all_args.pop(0)
    #   else:
    #     next_arg = None
    #   print(arg, next_arg, kwargs)

    #   if arg.startswith('-'):
    #     # this is a flag, this can be of two types one with = in it and one without
    #     if '=' in arg:
    #       # this is a flag with = in it
    #       key, value = arg.split('=')
    #       kwargs[key.strip('-')] = value
    #     else:
    #       arg = arg.strip('-')
    #       if next_arg.startswith('-'):
    #         kwargs[arg] = True # boolean
    #       else:
    #         kwargs[arg] = next_arg
    #       if len(all_args):
    #         arg = all_args.pop(0)
    #       else:
    #         break
    #       continue
    #   else:
    #     if kwargs:
    #       # this means we have a positional argument and a keyword argument which can either be a mistake
    #       # or an object being initialised and then its attributes being set used
    #       try:
    #         comp = self._init_comp(comp, *args_list, **kwargs)
    #         comp = getattr(comp, arg, None)
    #         if comp is None:
    #           lines = [
    #             f"{TC.BOLD}{TC.FAIL}ERROR:{TC.ENDC} Could not find command: '{arg}'. Available commands are:\n"
    #           ] + [f"  {x}" for x in tuple(comp.__dir__()) if not x.startswith('_')]
    #           self._print_and_exit(lines)
            
    #         # a fresh args kwargs
    #         args_list = []
    #         kwargs = {}
    #       except:
    #         lines = [
    #           f"{TC.BOLD}{TC.FAIL}ERROR:{TC.ENDC} Could not find command: '{arg}'. Available commands are:\n"
    #         ] + [f"  {x}" for x in tuple(comp.__dir__()) if not x.startswith('_')]
    #         self._print_and_exit(lines)
    #     else:
    #       args_list.append(arg)
    #     # arg = all_args.pop(0)

    #   arg = next_arg
    
    # print(kwargs, args_list, arg, all_args)

    # comp(*args_list, **kwargs)

  def _print_and_exit(self, lines):
    text = '\n'.join(lines) + '\n'
    # More(text, out=sys.stderr)
    print(text, file=sys.stderr)
    sys.exit(1)

  def _init_comp(self, comp, *args, **kwargs):
    return comp(*args, **kwargs)

  def More(self, contents: str, out):
    """Run a user specified pager or fall back to the internal pager.

    Args:
      contents: The entire contents of the text lines to page.
      out: The output stream.
      prompt: The page break prompt.
      check_pager: Checks the PAGER env var and uses it if True.
    """
    import signal, subprocess

    pager = encoding.GetEncodedValue(os.environ, 'PAGER', None)
    if pager == '-':
      # Use the fallback Pager.
      pager = None
    elif not pager:
      # Search for a pager that handles ANSI escapes.
      for command in ('less', 'pager'):
        if files.FindExecutableOnPath(command):
          pager = command
          break
    if pager:
      # If the pager is less(1) then instruct it to display raw ANSI escape
      # sequences to enable colors and font embellishments.
      less_orig = encoding.GetEncodedValue(os.environ, 'LESS', None)
      less = '-R' + (less_orig or '')
      encoding.SetEncodedValue(os.environ, 'LESS', less)
      # Ignore SIGINT while the pager is running.
      # We don't want to terminate the parent while the child is still alive.
      signal.signal(signal.SIGINT, signal.SIG_IGN)
      p = subprocess.Popen(pager, stdin=subprocess.PIPE, shell=True)
      enc = console_attr.GetConsoleAttr().GetEncoding()
      p.communicate(input=contents.encode(enc))
      p.wait()
      # Start using default signal handling for SIGINT again.
      signal.signal(signal.SIGINT, signal.SIG_DFL)
      if less_orig is None:
        encoding.SetEncodedValue(os.environ, 'LESS', None)
      return
    else:
      out.write(contents)




"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import proto.relics_pb2
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class CreateRelicRequest(google.protobuf.message.Message):
    """things related to general WebAPIs"""
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    WORKSPACE_ID_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    REGION_FIELD_NUMBER: builtins.int
    NBX_INTEGRATION_TOKEN_FIELD_NUMBER: builtins.int
    NBX_RESOURCE_ID_FIELD_NUMBER: builtins.int
    BUCKET_META_FIELD_NUMBER: builtins.int
    workspace_id: typing.Text
    """the main separation layer"""

    name: typing.Text
    region: typing.Text
    nbx_integration_token: typing.Text
    """deprecated, will be removed soon"""

    nbx_resource_id: typing.Text
    """this is the resource id for which will be paired with integration token"""

    @property
    def bucket_meta(self) -> proto.relics_pb2.BucketMetadata:
        """If the bucket is in their platform then they tell us that they want one there
        else we are going to assume that they want one in our platform
        """
        pass
    def __init__(self,
        *,
        workspace_id: typing.Text = ...,
        name: typing.Text = ...,
        region: typing.Text = ...,
        nbx_integration_token: typing.Text = ...,
        nbx_resource_id: typing.Text = ...,
        bucket_meta: typing.Optional[proto.relics_pb2.BucketMetadata] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["bucket_meta",b"bucket_meta"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["bucket_meta",b"bucket_meta","name",b"name","nbx_integration_token",b"nbx_integration_token","nbx_resource_id",b"nbx_resource_id","region",b"region","workspace_id",b"workspace_id"]) -> None: ...
global___CreateRelicRequest = CreateRelicRequest

class ListRelicsRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    WORKSPACE_ID_FIELD_NUMBER: builtins.int
    RELIC_NAME_FIELD_NUMBER: builtins.int
    PAGE_NO_FIELD_NUMBER: builtins.int
    RELIC_ID_FIELD_NUMBER: builtins.int
    workspace_id: typing.Text
    """the main separation layer"""

    relic_name: typing.Text
    """the name of the relic"""

    page_no: builtins.int
    """the page number for pagination needs"""

    relic_id: typing.Text
    """the ID of the parent relic"""

    def __init__(self,
        *,
        workspace_id: typing.Text = ...,
        relic_name: typing.Text = ...,
        page_no: builtins.int = ...,
        relic_id: typing.Text = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["page_no",b"page_no","relic_id",b"relic_id","relic_name",b"relic_name","workspace_id",b"workspace_id"]) -> None: ...
global___ListRelicsRequest = ListRelicsRequest

class ListRelicsResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    RELICS_FIELD_NUMBER: builtins.int
    TOTAL_RELICS_FIELD_NUMBER: builtins.int
    @property
    def relics(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[proto.relics_pb2.Relic]: ...
    total_relics: builtins.int
    """the total number of relics used for pagination"""

    def __init__(self,
        *,
        relics: typing.Optional[typing.Iterable[proto.relics_pb2.Relic]] = ...,
        total_relics: builtins.int = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["relics",b"relics","total_relics",b"total_relics"]) -> None: ...
global___ListRelicsResponse = ListRelicsResponse

class ListRelicFilesRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    WORKSPACE_ID_FIELD_NUMBER: builtins.int
    RELIC_ID_FIELD_NUMBER: builtins.int
    RELIC_NAME_FIELD_NUMBER: builtins.int
    PREFIX_FIELD_NUMBER: builtins.int
    FILE_NAME_FIELD_NUMBER: builtins.int
    PAGE_NO_FIELD_NUMBER: builtins.int
    workspace_id: typing.Text
    """the main separation layer"""

    relic_id: typing.Text
    """the id of the relic"""

    relic_name: typing.Text
    """the id of the relic"""

    prefix: typing.Text
    """the prefix for the files"""

    file_name: typing.Text
    """for searching through the files, we may not support this functionality
    right now, but this is a crucial system to have
    """

    page_no: builtins.int
    """the page number for pagination needs"""

    def __init__(self,
        *,
        workspace_id: typing.Text = ...,
        relic_id: typing.Text = ...,
        relic_name: typing.Text = ...,
        prefix: typing.Text = ...,
        file_name: typing.Text = ...,
        page_no: builtins.int = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["file_name",b"file_name","page_no",b"page_no","prefix",b"prefix","relic_id",b"relic_id","relic_name",b"relic_name","workspace_id",b"workspace_id"]) -> None: ...
global___ListRelicFilesRequest = ListRelicFilesRequest

class ListRelicFilesResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    FILES_FIELD_NUMBER: builtins.int
    TOTAL_FILES_FIELD_NUMBER: builtins.int
    @property
    def files(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[proto.relics_pb2.RelicFile]: ...
    total_files: builtins.int
    """the total number of files used for pagination"""

    def __init__(self,
        *,
        files: typing.Optional[typing.Iterable[proto.relics_pb2.RelicFile]] = ...,
        total_files: builtins.int = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["files",b"files","total_files",b"total_files"]) -> None: ...
global___ListRelicFilesResponse = ListRelicFilesResponse



from abc import abstractmethod
from typing import List

class BaseStore(object):
  """This is the base class for all Relic Stores. Instead of treating these functions as just means to an
  objective, we treat them as a ideas to organise things."""
  @abstractmethod
  def put(self, local_path: str) -> None:
    """Puts whatever you have at local_path to the same path in the store"""
    pass

  @abstractmethod
  def put_object(self, local_path: str, py_object) -> None:
    """Convinience function for putting a python object in the store"""
    pass

  @abstractmethod
  def put_to(self, local_path: str, remote_path: str) -> None:
    """like put but with a different remote path"""
    pass

  @abstractmethod
  def list_files(self, local_path: str) -> List[str]:
    """list all files in this relic"""
    pass

  @abstractmethod
  def has(self, local_path: str) -> bool:
    """check if this file exists in this relic"""
    pass

  @abstractmethod
  def get(self, local_path: str) -> None:
    """gets the put-ed file from the store to local_path"""
    pass

  @abstractmethod
  def get_object(self, local_path: str) -> None:
    """Convinience function for getting a python object from the store"""
    pass

  @abstractmethod
  def get_from(self, local_path: str, remote_path: str) -> None:
    """like get but with a different remote path"""
    pass

  @abstractmethod
  def rm(self, local_path: str) -> None:
    """delete this file in this relic, should not have side efects, meaning relic should only delete the file
    in the areas of its control. ex: If a file is ``put`` on relics, deleting it should not delete the file
    locally."""
    pass

  @abstractmethod
  def delete(self) -> None:
    """delete the relic"""
    pass



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/relics.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x12proto/relics.proto\"\xa7\x01\n\x05Relic\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x12\n\ncreated_on\x18\x03 \x01(\x03\x12\x15\n\rlast_modified\x18\x04 \x01(\x03\x12\x0f\n\x07starred\x18\x05 \x01(\x08\x12\x0c\n\x04tags\x18\x06 \x03(\t\x12\x14\n\x0cworkspace_id\x18\x07 \x01(\t\x12$\n\x0b\x62ucket_meta\x18\x64 \x01(\x0b\x32\x0f.BucketMetadata\"\xa2\x02\n\x0e\x42ucketMetadata\x12\x13\n\x0b\x62ucket_name\x18\x01 \x01(\t\x12\x0e\n\x06region\x18\x02 \x01(\t\x12(\n\x07\x62\x61\x63kend\x18\x03 \x01(\x0e\x32\x17.BucketMetadata.Backend\x12\x34\n\x0b\x62ucket_tags\x18\x04 \x03(\x0b\x32\x1f.BucketMetadata.BucketTagsEntry\x1a\x31\n\x0f\x42ucketTagsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"X\n\x07\x42\x61\x63kend\x12\t\n\x05UNSET\x10\x00\x12\n\n\x06\x41WS_S3\x10\x01\x12\x0b\n\x07GCP_GCS\x10\x02\x12\x0e\n\nAZURE_BLOB\x10\x03\x12\n\n\x06OCI_OB\x10\x04\x12\r\n\tDO_SPACES\x10\x05\"\xda\x03\n\tRelicFile\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x12\n\ncreated_on\x18\x02 \x01(\x03\x12\x15\n\rlast_modified\x18\x03 \x01(\x03\x12\x0f\n\x07starred\x18\x04 \x01(\x08\x12\x0c\n\x04tags\x18\x05 \x03(\t\x12\x0c\n\x04size\x18\x06 \x01(\x03\x12\x0f\n\x07\x63omment\x18\x07 \x01(\t\x12\x10\n\x08username\x18\x08 \x01(\t\x12\"\n\x04type\x18\t \x01(\x0e\x32\x14.RelicFile.RelicType\x12\x14\n\x0cworkspace_id\x18\n \x01(\t\x12\x12\n\nrelic_name\x18\x0b \x01(\t\x12\x10\n\x08relic_id\x18\x0c \x01(\t\x12\x0b\n\x03url\x18\x64 \x01(\t\x12(\n\x07headers\x18\x65 \x03(\x0b\x32\x17.RelicFile.HeadersEntry\x12\"\n\x04\x62ody\x18\x66 \x03(\x0b\x32\x14.RelicFile.BodyEntry\x1a.\n\x0cHeadersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a+\n\tBodyEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\",\n\tRelicType\x12\x08\n\x04\x46ILE\x10\x00\x12\n\n\x06\x46OLDER\x10\x01\x12\t\n\x05RELIC\x10\x02\x62\x06proto3')



_RELIC = DESCRIPTOR.message_types_by_name['Relic']
_BUCKETMETADATA = DESCRIPTOR.message_types_by_name['BucketMetadata']
_BUCKETMETADATA_BUCKETTAGSENTRY = _BUCKETMETADATA.nested_types_by_name['BucketTagsEntry']
_RELICFILE = DESCRIPTOR.message_types_by_name['RelicFile']
_RELICFILE_HEADERSENTRY = _RELICFILE.nested_types_by_name['HeadersEntry']
_RELICFILE_BODYENTRY = _RELICFILE.nested_types_by_name['BodyEntry']
_BUCKETMETADATA_BACKEND = _BUCKETMETADATA.enum_types_by_name['Backend']
_RELICFILE_RELICTYPE = _RELICFILE.enum_types_by_name['RelicType']
Relic = _reflection.GeneratedProtocolMessageType('Relic', (_message.Message,), {
  'DESCRIPTOR' : _RELIC,
  '__module__' : 'proto.relics_pb2'
  # @@protoc_insertion_point(class_scope:Relic)
  })
_sym_db.RegisterMessage(Relic)

BucketMetadata = _reflection.GeneratedProtocolMessageType('BucketMetadata', (_message.Message,), {

  'BucketTagsEntry' : _reflection.GeneratedProtocolMessageType('BucketTagsEntry', (_message.Message,), {
    'DESCRIPTOR' : _BUCKETMETADATA_BUCKETTAGSENTRY,
    '__module__' : 'proto.relics_pb2'
    # @@protoc_insertion_point(class_scope:BucketMetadata.BucketTagsEntry)
    })
  ,
  'DESCRIPTOR' : _BUCKETMETADATA,
  '__module__' : 'proto.relics_pb2'
  # @@protoc_insertion_point(class_scope:BucketMetadata)
  })
_sym_db.RegisterMessage(BucketMetadata)
_sym_db.RegisterMessage(BucketMetadata.BucketTagsEntry)

RelicFile = _reflection.GeneratedProtocolMessageType('RelicFile', (_message.Message,), {

  'HeadersEntry' : _reflection.GeneratedProtocolMessageType('HeadersEntry', (_message.Message,), {
    'DESCRIPTOR' : _RELICFILE_HEADERSENTRY,
    '__module__' : 'proto.relics_pb2'
    # @@protoc_insertion_point(class_scope:RelicFile.HeadersEntry)
    })
  ,

  'BodyEntry' : _reflection.GeneratedProtocolMessageType('BodyEntry', (_message.Message,), {
    'DESCRIPTOR' : _RELICFILE_BODYENTRY,
    '__module__' : 'proto.relics_pb2'
    # @@protoc_insertion_point(class_scope:RelicFile.BodyEntry)
    })
  ,
  'DESCRIPTOR' : _RELICFILE,
  '__module__' : 'proto.relics_pb2'
  # @@protoc_insertion_point(class_scope:RelicFile)
  })
_sym_db.RegisterMessage(RelicFile)
_sym_db.RegisterMessage(RelicFile.HeadersEntry)
_sym_db.RegisterMessage(RelicFile.BodyEntry)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _BUCKETMETADATA_BUCKETTAGSENTRY._options = None
  _BUCKETMETADATA_BUCKETTAGSENTRY._serialized_options = b'8\001'
  _RELICFILE_HEADERSENTRY._options = None
  _RELICFILE_HEADERSENTRY._serialized_options = b'8\001'
  _RELICFILE_BODYENTRY._options = None
  _RELICFILE_BODYENTRY._serialized_options = b'8\001'
  _RELIC._serialized_start=23
  _RELIC._serialized_end=190
  _BUCKETMETADATA._serialized_start=193
  _BUCKETMETADATA._serialized_end=483
  _BUCKETMETADATA_BUCKETTAGSENTRY._serialized_start=344
  _BUCKETMETADATA_BUCKETTAGSENTRY._serialized_end=393
  _BUCKETMETADATA_BACKEND._serialized_start=395
  _BUCKETMETADATA_BACKEND._serialized_end=483
  _RELICFILE._serialized_start=486
  _RELICFILE._serialized_end=960
  _RELICFILE_HEADERSENTRY._serialized_start=823
  _RELICFILE_HEADERSENTRY._serialized_end=869
  _RELICFILE_BODYENTRY._serialized_start=871
  _RELICFILE_BODYENTRY._serialized_end=914
  _RELICFILE_RELICTYPE._serialized_start=916
  _RELICFILE_RELICTYPE._serialized_end=960
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/relics_rpc.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from nbox.sublime.proto import relics_pb2 as proto_dot_relics__pb2
from nbox.sublime.proto import common_pb2 as proto_dot_common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x16proto/relics_rpc.proto\x1a\x12proto/relics.proto\x1a\x12proto/common.proto\"\xa6\x01\n\x12\x43reateRelicRequest\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0e\n\x06region\x18\x03 \x01(\t\x12\x1d\n\x15nbx_integration_token\x18\x04 \x01(\t\x12\x17\n\x0fnbx_resource_id\x18\x05 \x01(\t\x12$\n\x0b\x62ucket_meta\x18\x06 \x01(\x0b\x32\x0f.BucketMetadata\"`\n\x11ListRelicsRequest\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x12\n\nrelic_name\x18\x02 \x01(\t\x12\x0f\n\x07page_no\x18\x03 \x01(\x05\x12\x10\n\x08relic_id\x18\x04 \x01(\t\"B\n\x12ListRelicsResponse\x12\x16\n\x06relics\x18\x01 \x03(\x0b\x32\x06.Relic\x12\x14\n\x0ctotal_relics\x18\x02 \x01(\x05\"\x87\x01\n\x15ListRelicFilesRequest\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x10\n\x08relic_id\x18\x02 \x01(\t\x12\x12\n\nrelic_name\x18\x03 \x01(\t\x12\x0e\n\x06prefix\x18\x04 \x01(\t\x12\x11\n\tfile_name\x18\x05 \x01(\t\x12\x0f\n\x07page_no\x18\x06 \x01(\x05\"H\n\x16ListRelicFilesResponse\x12\x19\n\x05\x66iles\x18\x01 \x03(\x0b\x32\n.RelicFile\x12\x13\n\x0btotal_files\x18\x02 \x01(\x05\x32\xbd\x03\n\nRelicStore\x12-\n\x0c\x63reate_relic\x12\x13.CreateRelicRequest\x1a\x06.Relic\"\x00\x12\x38\n\x0blist_relics\x12\x12.ListRelicsRequest\x1a\x13.ListRelicsResponse\"\x00\x12+\n\x11update_relic_meta\x12\x06.Relic\x1a\x0c.Acknowledge\"\x00\x12&\n\x0c\x64\x65lete_relic\x12\x06.Relic\x1a\x0c.Acknowledge\"\x00\x12%\n\x11get_relic_details\x12\x06.Relic\x1a\x06.Relic\"\x00\x12\'\n\x0b\x63reate_file\x12\n.RelicFile\x1a\n.RelicFile\"\x00\x12\x45\n\x10list_relic_files\x12\x16.ListRelicFilesRequest\x1a\x17.ListRelicFilesResponse\"\x00\x12/\n\x11\x64\x65lete_relic_file\x12\n.RelicFile\x1a\x0c.Acknowledge\"\x00\x12)\n\rdownload_file\x12\n.RelicFile\x1a\n.RelicFile\"\x00\x62\x06proto3')



_CREATERELICREQUEST = DESCRIPTOR.message_types_by_name['CreateRelicRequest']
_LISTRELICSREQUEST = DESCRIPTOR.message_types_by_name['ListRelicsRequest']
_LISTRELICSRESPONSE = DESCRIPTOR.message_types_by_name['ListRelicsResponse']
_LISTRELICFILESREQUEST = DESCRIPTOR.message_types_by_name['ListRelicFilesRequest']
_LISTRELICFILESRESPONSE = DESCRIPTOR.message_types_by_name['ListRelicFilesResponse']
CreateRelicRequest = _reflection.GeneratedProtocolMessageType('CreateRelicRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATERELICREQUEST,
  '__module__' : 'proto.relics_rpc_pb2'
  # @@protoc_insertion_point(class_scope:CreateRelicRequest)
  })
_sym_db.RegisterMessage(CreateRelicRequest)

ListRelicsRequest = _reflection.GeneratedProtocolMessageType('ListRelicsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTRELICSREQUEST,
  '__module__' : 'proto.relics_rpc_pb2'
  # @@protoc_insertion_point(class_scope:ListRelicsRequest)
  })
_sym_db.RegisterMessage(ListRelicsRequest)

ListRelicsResponse = _reflection.GeneratedProtocolMessageType('ListRelicsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTRELICSRESPONSE,
  '__module__' : 'proto.relics_rpc_pb2'
  # @@protoc_insertion_point(class_scope:ListRelicsResponse)
  })
_sym_db.RegisterMessage(ListRelicsResponse)

ListRelicFilesRequest = _reflection.GeneratedProtocolMessageType('ListRelicFilesRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTRELICFILESREQUEST,
  '__module__' : 'proto.relics_rpc_pb2'
  # @@protoc_insertion_point(class_scope:ListRelicFilesRequest)
  })
_sym_db.RegisterMessage(ListRelicFilesRequest)

ListRelicFilesResponse = _reflection.GeneratedProtocolMessageType('ListRelicFilesResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTRELICFILESRESPONSE,
  '__module__' : 'proto.relics_rpc_pb2'
  # @@protoc_insertion_point(class_scope:ListRelicFilesResponse)
  })
_sym_db.RegisterMessage(ListRelicFilesResponse)

_RELICSTORE = DESCRIPTOR.services_by_name['RelicStore']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _CREATERELICREQUEST._serialized_start=67
  _CREATERELICREQUEST._serialized_end=233
  _LISTRELICSREQUEST._serialized_start=235
  _LISTRELICSREQUEST._serialized_end=331
  _LISTRELICSRESPONSE._serialized_start=333
  _LISTRELICSRESPONSE._serialized_end=399
  _LISTRELICFILESREQUEST._serialized_start=402
  _LISTRELICFILESREQUEST._serialized_end=537
  _LISTRELICFILESRESPONSE._serialized_start=539
  _LISTRELICFILESRESPONSE._serialized_end=611
  _RELICSTORE._serialized_start=614
  _RELICSTORE._serialized_end=1059
# @@protoc_insertion_point(module_scope)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import abc
import collections.abc
import google.protobuf.empty_pb2
import grpc
import serve_pb2

class ServingServiceStub:
    def __init__(self, channel: grpc.Channel) -> None: ...
    Create: grpc.UnaryUnaryMultiCallable[
        serve_pb2.ServingRequest,
        serve_pb2.Serving,
    ]
    """Creates a new Serving Group.
    If `.models` are specified, the Serving Group will be created with new
    `len(.models)` models.
    """
    GetServing: grpc.UnaryUnaryMultiCallable[
        serve_pb2.ServingRequest,
        serve_pb2.Serving,
    ]
    """Get all available info for a Serving Group. All `.models` are populated."""
    ListServings: grpc.UnaryUnaryMultiCallable[
        serve_pb2.ServingListRequest,
        serve_pb2.ServingListResponse,
    ]
    """Get list of Serving Grpup(s) of a user/workspace."""
    DeleteServing: grpc.UnaryUnaryMultiCallable[
        serve_pb2.ServingRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    """Delete a Serving Group and all its models."""
    UpdateServing: grpc.UnaryUnaryMultiCallable[
        serve_pb2.UpdateServingRequest,
        serve_pb2.Serving,
    ]
    """Update a Serving Group. Only the fields specified in the `update_mask` are
    updated.
    """

class ServingServiceServicer(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def Create(
        self,
        request: serve_pb2.ServingRequest,
        context: grpc.ServicerContext,
    ) -> serve_pb2.Serving:
        """Creates a new Serving Group.
        If `.models` are specified, the Serving Group will be created with new
        `len(.models)` models.
        """
    @abc.abstractmethod
    def GetServing(
        self,
        request: serve_pb2.ServingRequest,
        context: grpc.ServicerContext,
    ) -> serve_pb2.Serving:
        """Get all available info for a Serving Group. All `.models` are populated."""
    @abc.abstractmethod
    def ListServings(
        self,
        request: serve_pb2.ServingListRequest,
        context: grpc.ServicerContext,
    ) -> serve_pb2.ServingListResponse:
        """Get list of Serving Grpup(s) of a user/workspace."""
    @abc.abstractmethod
    def DeleteServing(
        self,
        request: serve_pb2.ServingRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty:
        """Delete a Serving Group and all its models."""
    @abc.abstractmethod
    def UpdateServing(
        self,
        request: serve_pb2.UpdateServingRequest,
        context: grpc.ServicerContext,
    ) -> serve_pb2.Serving:
        """Update a Serving Group. Only the fields specified in the `update_mask` are
        updated.
        """

def add_ServingServiceServicer_to_server(servicer: ServingServiceServicer, server: grpc.Server) -> None: ...

class ModelServiceStub:
    def __init__(self, channel: grpc.Channel) -> None: ...
    Watch: grpc.UnaryStreamMultiCallable[
        serve_pb2.ModelRequest,
        serve_pb2.ServingStatus,
    ]
    """Watch for status update of a model in Serving Group."""
    UploadModel: grpc.UnaryUnaryMultiCallable[
        serve_pb2.ModelRequest,
        serve_pb2.Model,
    ]
    """Get S3 Upload URL for new/existing Model in Serving Group.
    If id is none then creates new Model with id `id` in the Serving Group.
    If model's serving_group_id is none/doesn't exist, then throws error.
    """
    Deploy: grpc.UnaryUnaryMultiCallable[
        serve_pb2.ModelRequest,
        serve_pb2.Model,
    ]
    """Deploy new model in Serving Group."""
    ModelLogs: grpc.UnaryStreamMultiCallable[
        serve_pb2.ModelRequest,
        serve_pb2.ModelLog,
    ]
    """Get logs of a model."""
    ListModels: grpc.UnaryUnaryMultiCallable[
        serve_pb2.ModelListRequest,
        serve_pb2.ModelListResponse,
    ]
    """Get list of models in a Serving Group."""
    DeleteModel: grpc.UnaryUnaryMultiCallable[
        serve_pb2.ModelRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    """Delete a model in a Serving Group."""
    SetModelPin: grpc.UnaryUnaryMultiCallable[
        serve_pb2.ModelRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    """Pinned a model of a Serving Group."""

class ModelServiceServicer(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def Watch(
        self,
        request: serve_pb2.ModelRequest,
        context: grpc.ServicerContext,
    ) -> collections.abc.Iterator[serve_pb2.ServingStatus]:
        """Watch for status update of a model in Serving Group."""
    @abc.abstractmethod
    def UploadModel(
        self,
        request: serve_pb2.ModelRequest,
        context: grpc.ServicerContext,
    ) -> serve_pb2.Model:
        """Get S3 Upload URL for new/existing Model in Serving Group.
        If id is none then creates new Model with id `id` in the Serving Group.
        If model's serving_group_id is none/doesn't exist, then throws error.
        """
    @abc.abstractmethod
    def Deploy(
        self,
        request: serve_pb2.ModelRequest,
        context: grpc.ServicerContext,
    ) -> serve_pb2.Model:
        """Deploy new model in Serving Group."""
    @abc.abstractmethod
    def ModelLogs(
        self,
        request: serve_pb2.ModelRequest,
        context: grpc.ServicerContext,
    ) -> collections.abc.Iterator[serve_pb2.ModelLog]:
        """Get logs of a model."""
    @abc.abstractmethod
    def ListModels(
        self,
        request: serve_pb2.ModelListRequest,
        context: grpc.ServicerContext,
    ) -> serve_pb2.ModelListResponse:
        """Get list of models in a Serving Group."""
    @abc.abstractmethod
    def DeleteModel(
        self,
        request: serve_pb2.ModelRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty:
        """Delete a model in a Serving Group."""
    @abc.abstractmethod
    def SetModelPin(
        self,
        request: serve_pb2.ModelRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty:
        """Pinned a model of a Serving Group."""

def add_ModelServiceServicer_to_server(servicer: ModelServiceServicer, server: grpc.Server) -> None: ...



import sys
from collections.abc import Iterable
from functools import reduce

def _if(*args):
  assert len(args) >= 3, f"Provide atleast (cond, _true, _false) syntax "
  assert len(args) % 2 == 1, f"There should be odd elements in the conditional, got {len(args)}"
  pairs = [args[i*2:i*2+2] for i, _ in enumerate(args[:-2:2])]
  pairs.append((True, args[-1]))
  for c,o in pairs:
    if c:
      return o

def _merge(*args):
  # takes a bunch of elements and translates them into a 1D Array
  # find and past solution from SO! don't use your brains on this
  array = []
  for a in args:
    if isinstance(a, Iterable) and not isinstance(a, str):
      for i in _merge(*a):
        array.append(i)
    else:
      array.append(a)
  return array


class JsonLogic():
  def __init__(self, rules, data = {}):
    """
    Modified for our needs from: https://github.com/nadirizr/json-logic-py
    This is a Python implementation of the following jsonLogic JS library: https://github.com/jwadhams/json-logic-js

    Args:
      rules (dict): The rules to be applied
      data (dict, optional): The data to be used in the rules. Defaults to None.
    """
    self.rules = rules
    self.data = data
    self.operations = {
      # accessing data
      "var" : (
        lambda a, not_found = None: reduce(
          lambda data, key: (
            data.get(key, not_found) if type(data) == dict else data[int(key)]
            if (type(data) in [list, tuple] and str(key).lstrip("-").isdigit())
            else not_found
          ),
          str(a).split("."),
        data
        )
      ),
      # "missing",
      # "missing_some",

      # logic and boolean operators
      "if": _if,
      "=="  : (lambda a, b: a == b),
      "===" : (lambda a, b: a is b),
      "!="  : (lambda a, b: a != b),
      "!==" : (lambda a, b: a is not b),
      "!"   : (lambda a: not a),
      # "!!",
      "or"  : (lambda *args: reduce(lambda total, arg: total or arg, args, False)),
      "and" : (lambda *args: reduce(lambda total, arg: total and arg, args, True)),

      # numeric operators
      ">"   : (lambda a, b: a > b),
      ">="  : (lambda a, b: a >= b),
      "<"   : (lambda a, b, c = None:  a < b if (c is None) else (a < b) and (b < c)),
      "<="  : (lambda a, b, c = None: a <= b if (c is None) else (a <= b) and (b <= c)),
      # between
      "min" : (lambda *args: min(args)),
      "max" : (lambda *args: max(args)),
      "+" : (lambda *args: reduce(lambda total, arg: total + float(arg), args, 0.0)),
      "*" : (lambda *args: reduce(lambda total, arg: total * float(arg), args, 1.0)),
      "-" : (lambda a, b=None: -a if b is None else a - b),
      "/" : (lambda a, b=None: a if b is None else float(a) / float(b)),
      "%"   : (lambda a, b: a % b),

      # Array opertors
      # "map",
      # "reduce",
      # "filter",
      # "all",
      # "none",
      # "some",
      "merge": _merge,
      "in"  : (lambda a, b: a in b if "__contains__" in dir(b) else False),

      # string operators
      # "in" defined above
      "cat" : (lambda *args: "".join(args)),
      "substr": (lambda a, b, c=None: a[b:c] if c else a[b:]),
      
      # Misc.
      "log" : (lambda a: a if sys.stdout.write(str(a)) else a),
      "count": (lambda *args: sum(1 if a else 0 for a in args)),
      "?:"  : (lambda a, b, c: b if a else c),
    }
  
  def apply(self):
    tests = self.rules
    data = self.data

    # You've recursed to a primitive, stop!
    if tests is None or type(tests) != dict:
      return tests

    op = tuple(tests.keys())[0]
    values = tests[op]
    
    if op not in self.operations:
      raise RuntimeError(f"Unrecognized operation '{op}'")

    # Easy syntax for unary operators, like {"var": "x"} instead of strict {"var": ["x"]}
    if type(values) not in [list, tuple]:
      values = [values]

    # Recursion!
    values = map(lambda val: JsonLogic(val, data).apply(), values)

    return self.operations[op](*values)



"""
This is the hidden closet with all messy and cool things.

This submodule has files for managing the adapters for different frameworks. The important
repositories with the logic start with ``on_``, why?

#. ``on_ml``: contains the code for exporting and sharing the weights across different\
    ml libraries like pytorch, tensorflow, haiku, etc.
#. ``on_operators``: contains the code for exporting and importing operators from frameworks\
    like Airflow, Prefect, etc.
#. ``on_functions``: This library contains the code for static parsing of python and\
    creating the NBX-JobFlow
"""

# if you like what you see and want to work on more things like this, reach out research@nimblebox.ai

from .on_operators import *
